import os

import numpy as np
from astropy import units as u
from astropy.constants import c as cc
from astropy.table import Table
from scipy.interpolate import UnivariateSpline

from specreduce.compat import Spectrum
from specreduce.core import SpecreduceOperation


__all__ = ['FluxCalibration']


class FluxCalibration(SpecreduceOperation):
    """
    Carries out routine flux calibration operations.

    Parameters
    ----------
    object_spectrum : a Spectrum object
        The observed object spectrum to apply the sensfunc to,
        with the wavelength of the data points in Angstroms
        as the ``spectral_axis``, and the magnitudes of the
        data as the ``flux``.
    airmass : float
        The value of the airmass. Note: NOT the header keyword.
    zeropoint : float, optional
        Conversion factor for mag->flux. (Default is 48.60).

    """

    def __call__(self, object_spectrum, airmass=1.00, zeropoint=48.60):
        self.object_spectrum = object_spectrum
        self.airmass = airmass
        self.zeropoint = zeropoint

    def mag2flux(self, spec_in=None):
        """
        Convert magnitudes to flux units. This is important for dealing with standards
        and files from IRAF, which are stored in AB mag units. To be clear, this converts
        to "PHOTFLAM" units in IRAF-speak. Assumes the common flux zeropoint used in IRAF.

        Parameters
        ----------
        spec_in : a Spectrum object, optional
            An input spectrum with wavelength of the data points in Angstroms
            as the ``spectral_axis`` and magnitudes of the data as the ``flux``.

        Returns
        -------
        spec_out : specutils.Spectrum
            Containing both ``flux`` and ``spectral_axis`` data in which
            the ``flux`` has been properly converted from mag->flux.

        """
        if spec_in is None:
            spec_in = self.object_spectrum

        lamb = spec_in.spectral_axis
        mag = spec_in.flux

        flux = (10.0**((mag + self.zeropt) / (-2.5))) * (cc.to('AA/s').value / lamb ** 2.0)
        flux = flux * u.erg / u.s / u.angstrom / (u.cm * u.cm)

        spec_out = Spectrum(spectral_axis=lamb, flux=flux)

        return spec_out

    @staticmethod
    def obs_extinction(obs_file):
        """
        Load the observatory-specific airmass extinction file from the supplied library

        Parameters
        ----------
        obs_file : str, {'apoextinct.dat', 'ctioextinct.dat', 'kpnoextinct.dat', 'ormextinct.dat'}
            The observatory-specific airmass extinction file. If not known for your
            observatory, use one of the provided files (e.g. `kpnoextinct.dat`).
            Following IRAF standard, extinction files have 2-column format
            wavelength (Angstroms), Extinction (Mag per Airmass).

        Returns
        -------
        Xfile: `~astropy.table.Table`
            Table with the observatory extinction data.

        """
        if len(obs_file) == 0:
            raise ValueError('Must select an observatory extinction file.')

        dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                           'datasets', 'extinction')

        if not os.path.isfile(os.path.join(dir, obs_file)):
            msg = "No valid standard star found at: " + os.path.join(dir, obs_file)
            raise ValueError(msg)

        # To read in the airmass extinction curve
        Xfile = Table.read(os.path.join(dir, obs_file), format='ascii', names=('wave', 'X'))
        Xfile['wave'].unit = 'AA'

        return Xfile

    def airmass_cor(self, Xfile):
        """
        Correct the spectrum based on the airmass.
        Requires observatory extinction file.

        Parameters
        ----------
        Xfile : `~astropy.table.Table`
            The extinction table from `obs_extinction`, with columns ('wave', 'X')
            that have standard units of: (angstroms, mag/airmass).

        Returns
        -------
        airmass_cor_spec : specutils.Spectrum
            The airmass-corrected Spectrum object.

        """
        object_spectrum = self.mag2flux()
        airmass = self.airmass

        obj_wave, obj_flux = object_spectrum.spectral_axis, object_spectrum.flux

        # linear interpol airmass extinction onto observed wavelengths
        new_X = np.interp(obj_wave.value, Xfile['wave'], Xfile['X'])

        # air_cor in units of mag/airmass, convert to flux/airmass
        airmass_ext = 10.0**(0.4 * airmass * new_X)

        airmass_cor_spec = Spectrum(flux=obj_flux * airmass_ext, spectral_axis=obj_wave)

        return airmass_cor_spec

    def onedstd(self, stdstar):
        """
        Load the onedstd from the supplied library.

        Parameters
        ----------
        stdstar : str
            Name of the standard star file in the specreduce/datasets/onedstds
            directory to be used for the flux calibration. The user must provide the
            subdirectory and file name.

            For example::

                standard_sensfunc(obj_wave, obj_flux, stdstar='spec50cal/bd284211.dat', mode='spline')

            If no std is supplied, or an improper path is given, raises a ValueError.

        Returns
        -------
        standard : `~astropy.table.Table`
            A table with the onedstd data.
        """  # noqa: E501
        std_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               'datasets', 'onedstds')

        if not os.path.isfile(os.path.join(std_dir, stdstar)):
            msg = "No valid standard star found at: " + os.path.join(std_dir, stdstar)
            raise ValueError(msg)

        standard = Table.read(os.path.join(std_dir, stdstar),
                              format='ascii', names=('wave', 'mag', 'width'))
        standard['wave'].unit = u.angstrom
        standard['width'].unit = u.angstrom

        # Standard star spectrum is stored in magnitude units (IRAF conventions)
        std_flux = self.mag2flux(spec_in=Spectrum(flux=standard['mag'],
                                                  spectral_axis=standard['wave']))
        std_flux = std_flux.flux
        standard['mag'].unit = u.mag
        standard.add_column(std_flux, name='flux')

        return standard

    def standard_sensfunc(self, standard, mode='linear', polydeg=9,
                          badlines=[6563, 4861, 4341], display=False):
        """
        Compute the standard star sensitivity function.

        Parameters
        ----------
        standard : `~astropy.table.Table`
            Output from ``onedstd``, has columns ('wave', 'width', 'mag', 'flux').
        mode : str, optional
            Can be "linear", "spline", or "poly". (Default is linear).
        polydeg : float, optional
            If mode='poly', this is the order of the polynomial to fit through.
            (Default is 9.)
        display : bool, optional
            If True, plot the sensfunc. (Default is False.)
            This requires ``matplotlib`` to be installed.
        badlines : array-like list
            A list of values (lines) to mask-out of when generating sensfunc.

        Returns
        -------
        sensfunc_spec : specutils.Spectrum
            The sensitivity function in the covered wavelength range
            for the given standard star.

        """
        spec = self.mag2flux()

        obj_wave, obj_flux = spec.spectral_axis, spec.flux

        # Automatically exclude some lines b/c resolution dependent response
        badlines = np.array(badlines, dtype='float')  # Balmer lines

        # Down-sample (ds) the observed flux to the standard's bins
        obj_flux_ds = np.array([], dtype=np.float)
        obj_wave_ds = np.array([], dtype=np.float)
        std_flux_ds = np.array([], dtype=np.float)
        for i in range(len(standard['flux'])):
            rng = np.where((obj_wave.value >= standard['wave'][i] - standard['width'][i] / 2.0) &
                           (obj_wave.value < standard['wave'][i] + standard['width'][i] / 2.0))[0]

            IsH = np.where((badlines >= standard['wave'][i] - standard['width'][i] / 2.0) &
                           (badlines < standard['wave'][i] + standard['width'][i] / 2.0))[0]

            # Does this bin contain observed spectra, and no Balmer lines?
            if (len(rng) > 1) and (len(IsH) == 0):
                obj_flux_ds = np.append(obj_flux_ds, np.nanmean(obj_flux.value[rng]))
                obj_wave_ds = np.append(obj_wave_ds, standard['wave'][i])
                std_flux_ds = np.append(std_flux_ds, standard['flux'][i])

        # the ratio between the standard star catalog flux and observed flux
        ratio = np.abs(std_flux_ds / obj_flux_ds)

        # The actual fit the log of this sensfunc ratio
        # Since IRAF does the 2.5*log(ratio), everything would be in mag units
        LogSensfunc = np.log10(ratio)

        # If invalid interpolation mode selected, make it spline
        if mode.lower() not in ('linear', 'spline', 'poly'):
            mode = 'spline'
            import warnings
            warnings.warn("WARNING: invalid mode set. Changing to default mode 'spline'")

        # Interpolate the calibration (sensfunc) on to observed wavelength grid
        if mode.lower() == 'linear':
            sensfunc2 = np.interp(obj_wave.value, obj_wave_ds, LogSensfunc)
        elif mode.lower() == 'spline':
            spl = UnivariateSpline(obj_wave_ds, LogSensfunc, ext=0, k=2, s=0.0025)
            sensfunc2 = spl(obj_wave.value)
        elif mode.lower() == 'poly':
            fit = np.polyfit(obj_wave_ds, LogSensfunc, polydeg)
            sensfunc2 = np.polyval(fit, obj_wave.value)

        sensfunc_out = (10 ** sensfunc2) * standard['flux'].unit / obj_flux.unit

        sensfunc_spec = Spectrum(spectral_axis=obj_wave, flux=sensfunc_out)

        if display is True:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(obj_wave, obj_flux * sensfunc_out, c="C0",
                     label="Observed x sensfunc", alpha=0.5)

            # plt.scatter(standard['wave'], std_flux, color='C1', alpha=0.75, label="stdstar")
            plt.scatter(obj_wave_ds, std_flux_ds, color='C1', alpha=0.75)

            plt.xlabel("Wavelength")
            plt.ylabel("Flux")

            plt.xlim(np.nanmin(obj_wave.value), np.nanmax(obj_wave.value))
            plt.ylim(np.nanmin(obj_flux.value * sensfunc_out.value) * 0.98,
                     np.nanmax(obj_flux.value * sensfunc_out.value) * 1.02)
            # plt.legend()
            plt.show()

        return sensfunc_spec

    def apply_sensfunc(self, sensfunc):
        """
        Apply the derived sensitivity function, converts observed units (e.g. ADU/s)
        to physical units (e.g., erg/s/cm2/A).
        Sensitivity function is first linearly interpolated onto the wavelength scale
        of the observed data, and then directly multiplied.

        Parameters
        ----------
        sensfunc : `~astropy.table.Table`
            The output of ``standard_sensfunc``, table has columns ('wave', 'S').

        Returns
        -------
        fluxcal_spec : specutils.Spectrum
            The sensfunc corrected ``Spectrum`` object.

        """
        spec = self.mag2flux()

        obj_wave, obj_flux = spec.spectral_axis, spec.flux

        # Sort, in case the sensfunc wavelength axis is backwards
        ss = np.argsort(obj_wave.value)

        # Interpolate the sensfunc onto the observed wavelength axis
        sensfunc2 = np.interp(obj_wave.value, sensfunc['wave'][ss], sensfunc['S'][ss])

        object_spectrum = obj_flux * (sensfunc2 * sensfunc['S'].unit)

        fluxcal_spec = Spectrum(spectral_axis=obj_wave, flux=object_spectrum)

        return fluxcal_spec



import numpy as np
import pytest
from astropy.wcs import WCS
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from specutils.fitting import fit_generic_continuum

from specreduce.calibration_data import load_pypeit_calibration_lines
from specreduce.compat import Spectrum
from specreduce.extract import BoxcarExtract
from specreduce.line_matching import match_lines_wcs, find_arc_lines
from specreduce.tracing import FlatTrace
from specreduce.utils.synth_data import make_2d_arc_image


@pytest.fixture
def mk_test_data():
    """
    Create test data for the line matching routines.
    """
    non_linear_header = {
        'CTYPE1': 'AWAV-GRA',  # Grating dispersion function with air wavelengths
        'CUNIT1': 'Angstrom',  # Dispersion units
        'CRPIX1': 519.8,       # Reference pixel [pix]
        'CRVAL1': 7245.2,      # Reference value [Angstrom]
        'CDELT1': 2.956,       # Linear dispersion [Angstrom/pix]
        'PV1_0': 4.5e5,        # Grating density [1/m]
        'PV1_1': 1,            # Diffraction order
        'PV1_2': 27.0,         # Incident angle [deg]
        'PV1_3': 1.765,        # Reference refraction
        'PV1_4': -1.077e6,     # Refraction derivative [1/m]
        'CTYPE2': 'PIXEL',     # Spatial detector coordinates
        'CUNIT2': 'pix',       # Spatial units
        'CRPIX2': 1,           # Reference pixel
        'CRVAL2': 0,           # Reference value
        'CDELT2': 1            # Spatial units per pixel
    }
    linear_header = {
        'CTYPE1': 'AWAV',  # Grating dispersion function with air wavelengths
        'CUNIT1': 'Angstrom',  # Dispersion units
        'CRPIX1': 519.8,       # Reference pixel [pix]
        'CRVAL1': 7245.2,      # Reference value [Angstrom]
        'CDELT1': 2.956,       # Linear dispersion [Angstrom/pix]
        'CTYPE2': 'PIXEL',     # Spatial detector coordinates
        'CUNIT2': 'pix',       # Spatial units
        'CRPIX2': 1,           # Reference pixel
        'CRVAL2': 0,           # Reference value
        'CDELT2': 1            # Spatial units per pixel
    }
    non_linear_wcs = WCS(header=non_linear_header)
    linear_wcs = WCS(header=linear_header)

    tilt_mod = models.Legendre1D(degree=2, c0=50, c1=0, c2=100)
    match_im = make_2d_arc_image(
        nx=1400,
        ny=1024,
        linelists=['HeI', 'NeI'],
        wcs=linear_wcs,
        line_fwhm=5,
        tilt_func=tilt_mod,
        amplitude_scale=5e-4
    )

    arclist = load_pypeit_calibration_lines(['HeI', 'NeI'])['wavelength']

    trace = FlatTrace(match_im, 512)
    arc_sp = BoxcarExtract(match_im, trace, width=5).spectrum
    arc_sp.uncertainty = StdDevUncertainty(np.sqrt(arc_sp.flux).value)
    continuum = fit_generic_continuum(arc_sp, median_window=51)
    arc_sub = Spectrum(
        spectral_axis=arc_sp.spectral_axis,
        flux=arc_sp.flux - continuum(arc_sp.spectral_axis)
    )
    arc_sub.uncertainty = arc_sp.uncertainty

    return linear_wcs, non_linear_wcs, arclist, arc_sub


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:No observer defined on WCS")
@pytest.mark.filterwarnings("ignore:Model is linear in parameters")
def test_find_arc_lines(mk_test_data):
    """
    Test the find_arc_lines routine.
    """
    _, _, _, arc_sub = mk_test_data
    lines = find_arc_lines(arc_sub, fwhm=5, window=5, noise_factor=5)
    assert len(lines) > 1


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:No observer defined on WCS")
@pytest.mark.filterwarnings("ignore:Model is linear in parameters")
def test_match_lines_wcs(mk_test_data):
    """
    Test the match_lines_wcs routine.
    """
    linear_wcs, _, arclist, arc_sub = mk_test_data
    lines = find_arc_lines(arc_sub, fwhm=5, window=5, noise_factor=5)
    matched_lines = match_lines_wcs(
        pixel_positions=lines['centroid'],
        catalog_wavelengths=arclist,
        spectral_wcs=linear_wcs.spectral,
        tolerance=5
    )
    assert len(matched_lines) > 1



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from specreduce.core import *  # noqa
from specreduce.wavelength_calibration import * # noqa

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''



import astropy.units as u
import numpy as np
import pytest
from astropy.nddata import NDData

from specreduce.background import Background
from specreduce.compat import Spectrum
from specreduce.tracing import FlatTrace, ArrayTrace


def test_background(
    mk_test_img_raw, mk_test_spec_no_spectral_axis, mk_test_spec_with_spectral_axis
):
    img = mk_test_img_raw
    image = mk_test_spec_no_spectral_axis
    image_um = mk_test_spec_with_spectral_axis
    #
    # Try combinations of extraction center, and even/odd
    # extraction aperture sizes.
    #
    trace_pos = 15
    trace = FlatTrace(image, trace_pos)
    bkg_sep = 5
    bkg_width = 2

    # all the following should be equivalent, whether image's spectral axis
    # is in pixels or physical units:
    bg1 = Background(image, [trace - bkg_sep, trace + bkg_sep], width=bkg_width)
    bg2 = Background.two_sided(image, trace, bkg_sep, width=bkg_width)
    bg3 = Background.two_sided(image, trace_pos, bkg_sep, width=bkg_width)
    assert np.allclose(bg1.bkg_image().flux, bg2.bkg_image().flux)
    assert np.allclose(bg1.bkg_image().flux, bg3.bkg_image().flux)

    bg4 = Background(image_um, [trace - bkg_sep, trace + bkg_sep], width=bkg_width)
    bg5 = Background.two_sided(image_um, trace, bkg_sep, width=bkg_width)
    bg6 = Background.two_sided(image_um, trace_pos, bkg_sep, width=bkg_width)
    assert np.allclose(bg1.bkg_image().flux, bg4.bkg_image().flux)
    assert np.allclose(bg1.bkg_image().flux, bg5.bkg_image().flux)
    assert np.allclose(bg1.bkg_image().flux, bg6.bkg_image().flux)

    # test that creating a one_sided background works
    Background.one_sided(image, trace, bkg_sep, width=bkg_width)

    # test that passing a single trace works
    bg = Background(image, trace, width=bkg_width)

    # test that image subtraction works
    sub1 = image - bg1
    sub2 = bg1.sub_image(image)
    sub3 = bg1.sub_image()
    assert np.allclose(sub1.flux, sub2.flux)
    assert np.allclose(sub2.flux, sub3.flux)

    sub4 = image_um - bg4
    sub5 = bg4.sub_image(image_um)
    sub6 = bg4.sub_image()
    assert np.allclose(sub1.flux, sub4.flux)
    assert np.allclose(sub4.flux, sub5.flux)
    assert np.allclose(sub5.flux, sub6.flux)

    bkg_spec = bg1.bkg_spectrum()
    assert isinstance(bkg_spec, Spectrum)
    sub_spec = bg1.sub_spectrum()
    assert isinstance(sub_spec, Spectrum)

    # test that width==0 results in no background
    bg = Background.two_sided(image, trace, bkg_sep, width=0)
    assert np.all(bg.bkg_image().flux == 0)

    # test that any NaNs in input image (whether in or outside the window) don't
    # propagate to _bkg_array (which affects bkg_image and sub_image methods) or
    # the final 1D spectra.
    img[0, 0] = np.nan  # out of window
    img[trace_pos, 0] = np.nan  # in window
    stats = ["average", "median"]

    for st in stats:
        bg = Background(img, trace - bkg_sep, width=bkg_width, statistic=st)
        assert np.isnan(bg.image.flux).sum() == 2
        assert np.isnan(bg._bkg_array).sum() == 0
        assert np.isnan(bg.bkg_spectrum().flux).sum() == 0
        assert np.isnan(bg.sub_spectrum().flux).sum() == 0

    with pytest.warns(DeprecationWarning, match="bkg_statistic.*deprecated"):
        bg.bkg_spectrum(bkg_statistic="mean")


def test_warnings_errors(mk_test_spec_no_spectral_axis):
    image = mk_test_spec_no_spectral_axis

    # image.shape (30, 10)
    with pytest.warns(match="background window extends beyond image boundaries"):
        Background.two_sided(image, 25, 4, width=3)

    # bottom of top window near/on top-edge of image (these should warn, but not fail)
    with pytest.warns(match="background window extends beyond image boundaries"):
        Background.two_sided(image, 25, 8, width=5)

    with pytest.warns(match="background window extends beyond image boundaries"):
        Background.two_sided(image, 25, 8, width=6)

    with pytest.warns(match="background window extends beyond image boundaries"):
        Background.two_sided(image, 25, 8, width=7)

    with pytest.warns(match="background window extends beyond image boundaries"):
        Background.two_sided(image, 7, 5, width=6)

    trace = ArrayTrace(image, trace=np.arange(10) + 20)  # from 20 to 29
    with pytest.warns(match="background window extends beyond image boundaries"):
        with pytest.raises(
            ValueError,
            match="background window does not remain in bounds across entire dispersion axis",
        ):  # noqa
            # 20 + 10 - 3 = 27 (lower edge of window on-image at right side of trace)
            # 29 + 10 - 3 = 36 (lower edge of window off-image at right side of trace)
            Background.one_sided(image, trace, 10, width=3)

    with pytest.raises(ValueError, match="width must be positive"):
        Background.two_sided(image, 25, 2, width=-1)


def test_trace_inputs(mk_test_img_raw):
    """
    Tests for the input argument 'traces' to `Background`. This should accept
    a list of or a single Trace object, or a list of or a single (positive)
    number to define a FlatTrace.
    """

    image = mk_test_img_raw

    # When `Background` object is created with no Trace object passed in it should
    # create a FlatTrace in the middle of the image (according to disp. axis)
    background = Background(image, width=5)
    assert np.all(background.traces[0].trace.data == image.shape[1] / 2.0)

    # FlatTrace(s) should be created if number or list of numbers is passed in for `traces`
    background = Background(image, 10.0, width=5)
    assert isinstance(background.traces[0], FlatTrace)
    assert background.traces[0].trace_pos == 10.0

    traces = [10.0, 15]
    background = Background(image, traces, width=5)
    for i, trace_pos in enumerate(traces):
        assert background.traces[i].trace_pos == trace_pos

    # make sure error is raised if input for `traces` is invalid
    match_str = (
        "objects, a number or list of numbers to define FlatTraces, "
        + "or None to use a FlatTrace in the middle of the image."
    )
    with pytest.raises(ValueError, match=match_str):
        Background(image, "non_valid_trace_pos")


class TestMasksBackground:
    """
    Various test functions to test how masked and non-finite data is handled
    in `Background.
    """

    def mk_img(self, nrows=4, ncols=5, nan_slices=None):
        """
        Make a simple gradient image to test masking in Background.
        Optionally add NaNs to data with `nan_slices`. Returned array is in
        u.DN.
        """

        img = np.tile((np.arange(1.0, ncols + 1)), (nrows, 1))

        if nan_slices:  # add nans in data
            for s in nan_slices:
                img[s] = np.nan

        return img * u.DN

    @pytest.mark.parametrize("mask", ["apply", "propagate", "zero_fill"])
    def test_fully_masked_column(self, mask):
        """
        Test background with some fully-masked columns (not fully masked image).
        In this case, the background value for that fully-masked column should
        be 0.0, with no error or warning raised.
        """

        img = self.mk_img(nrows=10, ncols=10)
        img[:, 0:1] = np.nan

        bkg = Background(img, traces=FlatTrace(img, 6), mask_treatment=mask)
        assert np.all(bkg.bkg_image().data[:, 0:1] == 0.0)

    @pytest.mark.parametrize("mask", ["apply", "propagate"])
    def test_fully_masked_image(self, mask):
        """
        Test that the appropriate error is raised by `Background` when image
        is fully masked/NaN.
        """

        with pytest.raises(ValueError, match="Image is fully masked."):
            # fully NaN image
            img = self.mk_img() * np.nan
            Background(img, traces=FlatTrace(self.mk_img(), 2), mask_treatment=mask)

        with pytest.raises(ValueError, match="Image is fully masked."):
            # fully masked image (should be equivalent)
            img = NDData(np.ones((4, 5)), mask=np.ones((4, 5), dtype=bool))
            Background(img, traces=FlatTrace(self.mk_img(), 2), mask_treatment=mask)

        # Now test that an image that isn't fully masked, but is fully masked
        # within the window determined by `width`, produces the correct result.
        msg = "Image is fully masked within background window determined by `width`."
        with pytest.raises(ValueError, match=msg):
            img = self.mk_img(nrows=12, ncols=12, nan_slices=[np.s_[3:10, :]])
            Background(img, traces=FlatTrace(img, 6), width=7)

    @pytest.mark.filterwarnings("ignore:background window extends beyond image boundaries")
    @pytest.mark.parametrize(
        "method,expected",
        [
            ("apply", np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])),
            (
                "propagate",
                np.array([0.0, 2.0, 3.0, 0.0, 5.0, 6.0, 7.0, 0.0, 9.0, 10.0, 11.0, 12.0]),
            ),
            (
                "zero_fill",
                np.array(
                    [
                        0.58333333,
                        2.0,
                        3.0,
                        2.33333333,
                        5.0,
                        6.0,
                        7.0,
                        7.33333333,
                        9.0,
                        10.0,
                        11.0,
                        12.0,
                    ]
                ),
            ),
        ],
    )
    def test_mask_treatment_bkg_img_spectrum(self, method, expected):
        """
        This test function tests `Background.bkg_image` and
        `Background.bkg_spectrum` when there is masked data. It also tests
        background subtracting the image, and returning the spectrum of the
        background subtracted image. This test is parameterized over all
        currently implemented mask handling methods to test that they
        work as intended. The window size is set to use the entire image array,
        so warning about background window is ignored."""

        img_size = 12  # square 12 x 12 image

        # make image, set some value to nan, which will be masked in the function
        image1 = self.mk_img(
            nrows=img_size, ncols=img_size, nan_slices=[np.s_[5:10, 0], np.s_[7:12, 3], np.s_[2, 7]]
        )

        # also make an image that doesn't have nonf data values, but has
        # masked values at the same locations, to make sure they give the same
        # results
        mask = ~np.isfinite(image1)
        dat = self.mk_img(nrows=img_size, ncols=img_size)
        image2 = NDData(dat, mask=mask)

        for image in [image1, image2]:

            # construct a flat trace in center of image
            trace = FlatTrace(image, img_size / 2)

            # create 'Background' object with `mask_treatment` set
            # 'width' should be > size of image to use all pix (but warning will
            # be raised, which we ignore.)
            background = Background(image, mask_treatment=method, traces=trace, width=img_size + 1)

            # test background image matches 'expected'
            bk_img = background.bkg_image()
            # change this and following assertions to assert_quantity_allclose once
            # issue #213 is fixed
            np.testing.assert_allclose(bk_img.flux.value, np.tile(expected, (img_size, 1)))

            # test background spectrum matches 'expected'
            bk_spec = background.bkg_spectrum()
            np.testing.assert_allclose(bk_spec.flux.value, expected)

    def test_sub_bkg_image(self):
        """
        Test that masked and non-finite data is handled correctly when subtracting
        background from image, for all currently implemented masking
        options.
        """

        # make image, set some value to nan, which will be masked in the function
        image = self.mk_img(
            nrows=12, ncols=12, nan_slices=[np.s_[5:10, 0], np.s_[7:12, 3], np.s_[2, 7]]
        )

        # Calculate a background value using mask_treatment = 'apply'.
        # For 'apply', the flag applies to how masked values are handled during
        # calculation of background for each column, but nonfinite data will
        # remain in input data array
        background_apply = Background(
            image, mask_treatment="apply", traces=FlatTrace(image, 6), width=2
        )
        subtracted_img_apply = background_apply.sub_image()

        assert np.all(np.isfinite(subtracted_img_apply.data) == np.isfinite(image.data))

        # Calculate a background value using mask_treatment = 'propagate'. The input
        # 2d mask is reduced to a 1d mask to mask out full columns in the
        # presence of any nans - this means that (as tested above in
        # `test_mask_treatment_bkg_img_spectrum`) those columns will have 0.0
        # background. In this case, image.mask is expanded to mask full
        # columns - the image itself will not have full columns set to np.nan,
        # so there are still valid background subtracted data values in this
        # case, but the corresponding mask for that entire column will be masked.

        background_propagate = Background(
            image, mask_treatment="propagate", traces=FlatTrace(image, 6), width=2
        )
        subtracted_img_propagate = background_propagate.sub_image()

        assert np.all(np.isfinite(subtracted_img_propagate.data) == np.isfinite(image.data))

        # Calculate a background value using mask_treatment = 'zero_fill'. Data
        # values at masked locations are set to 0 in the image array, and the
        # background value calculated for that column will be subtracted
        # resulting in a negative value. The resulting background subtracted
        # image should be fully finite and the mask should be zero everywhere
        # (all unmasked)

        background_zero_fill = Background(
            image, mask_treatment="zero_fill", traces=FlatTrace(image, 6), width=2
        )
        subtracted_img_zero_fill = background_zero_fill.sub_image()

        assert np.all(np.isfinite(subtracted_img_zero_fill.data))
        assert np.all(subtracted_img_zero_fill.mask == 0)



import pytest
from astropy import units as u
from astropy.modeling import models
from astropy.nddata import CCDData
from astropy.wcs import WCS

from specreduce.utils.synth_data import make_2d_trace_image, make_2d_arc_image, make_2d_spec_image


def test_make_2d_trace_image():
    ccdim = make_2d_trace_image(
        nx=3000,
        ny=1000,
        background=5,
        trace_center=None,
        trace_order=3,
        trace_coeffs={'c0': 0, 'c1': 50, 'c2': 100},
        profile=models.Gaussian1D(amplitude=100, stddev=10)
    )
    assert ccdim.data.shape == (1000, 3000)
    assert isinstance(ccdim, CCDData)


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:No observer defined on WCS")
def test_make_2d_arc_image_defaults():
    ccdim = make_2d_arc_image()
    assert isinstance(ccdim, CCDData)


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:No observer defined on WCS")
def test_make_2d_arc_pass_wcs():
    nx = 3000
    ny = 1000
    wave_unit = u.Angstrom
    extent = [3000, 6000]

    # test passing a valid WCS with dispersion along X
    wcs = WCS(naxis=2)
    wcs.wcs.ctype[0] = 'WAVE'
    wcs.wcs.ctype[1] = 'PIXEL'
    wcs.wcs.cunit[0] = wave_unit
    wcs.wcs.cunit[1] = u.pixel
    wcs.wcs.crval[0] = extent[0]
    wcs.wcs.cdelt[0] = (extent[1] - extent[0]) / nx
    wcs.wcs.crval[1] = 0
    wcs.wcs.cdelt[1] = 1

    ccdim = make_2d_arc_image(
        nx=nx,
        ny=ny,
        extent=None,
        wave_unit=None,
        wcs=wcs
    )
    assert ccdim.data.shape == (1000, 3000)
    assert isinstance(ccdim, CCDData)

    # test passing a tilt model
    tilt_model = models.Chebyshev1D(degree=2, c0=50, c1=0, c2=100)
    ccdim = make_2d_arc_image(
        nx=nx,
        ny=ny,
        extent=None,
        wave_unit=None,
        wcs=wcs,
        tilt_func=tilt_model
    )
    assert ccdim.data.shape == (1000, 3000)
    assert isinstance(ccdim, CCDData)

    # make sure WCS without spectral axis gets rejected
    wcs.wcs.ctype[0] = 'PIXEL'
    assert wcs.spectral.naxis == 0
    with pytest.raises(ValueError, match='Provided WCS must have a spectral axis'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=None,
            wave_unit=None,
            wcs=wcs
        )

    # test passing valid WCS with dispersion along Y while using air wavelengths
    wcs = WCS(naxis=2)
    wcs.wcs.ctype[1] = 'AWAV'
    wcs.wcs.ctype[0] = 'PIXEL'
    wcs.wcs.cunit[1] = wave_unit
    wcs.wcs.cunit[0] = u.pixel
    wcs.wcs.crval[1] = extent[0]
    wcs.wcs.cdelt[1] = (extent[1] - extent[0]) / nx
    wcs.wcs.crval[0] = 0
    wcs.wcs.cdelt[0] = 1

    ccdim = make_2d_arc_image(
        nx=ny,
        ny=nx,
        extent=None,
        wave_unit=None,
        wave_air=True,
        wcs=wcs,
        tilt_func=tilt_model
    )
    assert ccdim.data.shape == (3000, 1000)
    assert isinstance(ccdim, CCDData)

    # make sure no WCS and no extent gets rejected
    with pytest.raises(ValueError, match='Must specify either a wavelength extent or a WCS'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=None,
            wave_unit=None,
            wcs=None
        )

    # make sure if extent is provided, it has the right length
    with pytest.raises(ValueError, match='Wavelength extent must be of length 2'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=[1, 2, 3],
            wave_unit=None,
            wcs=None
        )

    # make sure a 1D WCS gets rejected
    wcs = WCS(naxis=1)
    wcs.wcs.ctype[0] = 'WAVE'
    wcs.wcs.cunit[0] = wave_unit
    wcs.wcs.crval[0] = extent[0]
    wcs.wcs.cdelt[0] = (extent[1] - extent[0]) / nx

    with pytest.raises(ValueError, match='WCS must have NAXIS=2 for a 2D image'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=None,
            wave_unit=None,
            wcs=wcs
        )

    # make sure a WCS with no spectral axis gets rejected
    wcs = WCS(naxis=2)
    wcs.wcs.ctype[1] = 'PIXEL'
    wcs.wcs.ctype[0] = 'PIXEL'
    wcs.wcs.cunit[1] = u.pixel
    wcs.wcs.cunit[0] = u.pixel
    wcs.wcs.crval[1] = extent[0]
    wcs.wcs.cdelt[1] = (extent[1] - extent[0]) / nx
    wcs.wcs.crval[0] = 0
    wcs.wcs.cdelt[0] = 1

    with pytest.raises(ValueError, match='Provided WCS must have a spectral axis'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=None,
            wave_unit=None,
            wcs=wcs
        )

    # make sure invalid wave_unit is caught
    with pytest.raises(ValueError, match='Wavelength unit must be a length unit'):
        ccdim = make_2d_arc_image(
            nx=nx,
            ny=ny,
            extent=[100, 300],
            wave_unit=u.pixel
        )

    # make sure a non-polynomial tilt_func gets rejected
    with pytest.raises(
        ValueError,
        match='The only tilt functions currently supported are 1D polynomials'
    ):
        ccdim = make_2d_arc_image(
            tilt_func=models.Gaussian1D
        )


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore:No observer defined on WCS")
def test_make_2d_spec_image_defaults():
    ccdim = make_2d_spec_image()
    assert isinstance(ccdim, CCDData)



import numpy as np
import pytest
from astropy import units as u
from astropy.modeling import models
from astropy.nddata import NDData, VarianceUncertainty, UnknownUncertainty, StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose

from specreduce.background import Background
from specreduce.compat import Spectrum
from specreduce.extract import BoxcarExtract, HorneExtract, OptimalExtract, _align_along_trace
from specreduce.tracing import FitTrace, FlatTrace, ArrayTrace


def add_gaussian_source(image, amps=2, stddevs=2, means=None):
    """Modify `image.data` to add a horizontal spectrum across the image.
    Each column can have a different amplitude, stddev or mean position
    if these are arrays (otherwise, constant across image).
    """

    nrows, ncols = image.shape

    if means is None:
        means = nrows // 2

    if not isinstance(means, np.ndarray):
        means = np.ones(ncols) * means
    if not isinstance(amps, np.ndarray):
        amps = np.ones(ncols) * amps
    if not isinstance(stddevs, np.ndarray):
        stddevs = np.ones(ncols) * stddevs

    for i, col in enumerate(range(ncols)):
        mod = models.Gaussian1D(amplitude=amps[i], mean=means[i], stddev=stddevs[i])
        image.data[:, i] = mod(np.arange(nrows))


def test_boxcar_extraction(mk_test_img):

    # Try combinations of extraction center, and even/odd extraction aperture sizes.
    image = mk_test_img

    trace = FlatTrace(image, 15.0)
    boxcar = BoxcarExtract(image, trace)

    spectrum = boxcar.spectrum
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 75.0))
    assert spectrum.unit is not None and spectrum.unit == u.Jy

    trace.set_position(14.5)
    spectrum = boxcar()
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 72.5))

    trace.set_position(14.7)
    spectrum = boxcar()
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 73.5))

    trace.set_position(15.0)
    boxcar.width = 6
    spectrum = boxcar()
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 90.0))

    trace.set_position(14.5)
    spectrum = boxcar(width=6)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 87.0))

    trace.set_position(15.0)
    spectrum = boxcar(width=4.5)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 67.5))

    trace.set_position(15.0)
    spectrum = boxcar(width=4.7)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 70.5))

    trace.set_position(14.3)
    spectrum = boxcar(width=4.7)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 67.15))


def test_boxcar_nonfinite_handling(mk_test_img):
    image = mk_test_img
    image.data[14, 2] = np.nan
    image.data[14, 4] = np.inf

    trace = FlatTrace(image, 15.0)
    spectrum = BoxcarExtract(image, trace, width=6, mask_treatment="ignore")()
    target = np.full_like(spectrum.flux.value, 90.0)
    target[2] = np.nan
    target[4] = np.inf
    np.testing.assert_equal(spectrum.flux.value, target)

    spectrum = BoxcarExtract(image, trace, width=6, mask_treatment="apply")()
    target[[2, 4]] = 91.2
    np.testing.assert_allclose(spectrum.flux.value, target, rtol=1e-8)


def test_boxcar_outside_image_condition(mk_test_img):

    # Trace is such that extraction aperture lays partially outside the image
    image = mk_test_img

    trace = FlatTrace(image, 3.0)
    boxcar = BoxcarExtract(image, trace, mask_treatment="apply")
    spectrum = boxcar(width=10.0)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 32.0))

    boxcar = BoxcarExtract(image, trace, mask_treatment="ignore")
    spectrum = boxcar(width=10.0)
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 32.0))


def test_boxcar_array_trace(mk_test_img):
    image = mk_test_img

    trace_array = np.ones_like(image[1]) * 15.0
    trace = ArrayTrace(image, trace_array)

    boxcar = BoxcarExtract(image, trace)
    spectrum = boxcar()
    assert np.allclose(spectrum.flux.value, np.full_like(spectrum.flux.value, 75.0))


def test_horne_image_validation(mk_test_img):
    #
    # Test HorneExtract scenarios specific to its use with an image of
    # type `~numpy.ndarray` (instead of the default `~astropy.nddata.NDData`).
    #
    image = mk_test_img

    trace = FlatTrace(image, 15.0)
    extract = OptimalExtract(image.data, trace)  # equivalent to HorneExtract

    # an array-type image must come with a variance argument
    with pytest.raises(ValueError, match=r".*variance must be specified.*"):
        ext = extract()

    # an NDData-type image can't have an empty uncertainty attribute
    with pytest.raises(ValueError, match=r".*NDData object lacks uncertainty"):
        ext = extract(image=image)

    # a warning should be raised if uncertainty is StdDevUncertainty
    with pytest.warns(UserWarning, match="image NDData object's uncertainty"):
        err = StdDevUncertainty(np.ones_like(image))
        image.uncertainty = err
        ext = extract(image=image)

    # an NDData-type image's uncertainty must be of type VarianceUncertainty
    # or type StdDevUncertainty
    with pytest.raises(ValueError, match=r".*unexpected uncertainty type.*"):
        err = UnknownUncertainty(np.ones_like(image))
        image.uncertainty = err
        ext = extract(image=image)

    # an array-type image must have the same dimensions as its variance argument
    with pytest.raises(ValueError, match=r".*shapes must match.*"):
        # remember variance, mask, and unit args are only checked if image
        # object doesn't have those attributes (e.g., numpy and Quantity arrays)
        err = np.ones_like(image[0])
        ext = extract(image=image.data, variance=err)

    # an array-type image must have the same dimensions as its mask argument
    with pytest.raises(ValueError, match=r".*shapes must match.*"):
        err = np.ones_like(image)
        mask = np.zeros_like(image[0])
        ext = extract(image=image.data, variance=err, mask=mask)

    # an array-type image given without mask and unit arguments is fine
    # and produces an extraction with flux in DN and spectral axis in pixels
    err = np.ones_like(image)
    ext = extract(image=image.data, variance=err, mask=None, unit=None)
    assert ext.unit == u.Unit("DN")
    assert np.all(ext.spectral_axis == np.arange(image.shape[extract.disp_axis]) * u.pix)


# ignore Astropy warning for extractions that aren't best fit with a Gaussian:
@pytest.mark.filterwarnings("ignore:The fit may be unsuccessful")
def test_horne_variance_errors(mk_test_img):
    image = mk_test_img

    trace = FlatTrace(image, 3.0)

    # all zeros are treated as non-weighted (i.e., as non-zero fluxes)
    image.uncertainty = VarianceUncertainty(np.zeros_like(image))
    image.mask = np.zeros_like(image)
    extract = HorneExtract(image, trace)
    ext = extract.spectrum
    assert not np.all(ext == 0)

    # single zero value adjusts mask and does not raise error
    err = np.ones_like(image)
    err[0][0] = 0
    image.uncertainty = VarianceUncertainty(err)
    ext = extract(image)
    assert not np.all(ext == 1)

    # single negative value raises error
    err = image.uncertainty.array
    err[0][0] = -1
    with pytest.raises(ValueError, match="variance must be fully positive"):
        # remember variance, mask, and unit args are only checked if image
        # object doesn't have those attributes (e.g., numpy and Quantity arrays)
        ext = extract(image=image.data, variance=err, mask=image.mask, unit=u.Jy)


@pytest.mark.filterwarnings("ignore:The fit may be unsuccessful")
def test_horne_non_flat_trace():
    # create a synthetic "2D spectrum" and its non-flat trace

    n_rows, n_cols = (10, 50)
    original = np.zeros((n_rows, n_cols))
    original[n_rows // 2] = 1

    # create small offsets along each column to specify a non-flat trace
    trace_offset = np.polyval([2e-3, -0.01, 0], np.arange(n_cols)).astype(int)
    exact_trace = n_rows // 2 - trace_offset

    # re-index the array with the offsets applied to the trace (make it non-flat):
    rows = np.broadcast_to(np.arange(n_rows)[:, None], original.shape)
    cols = np.broadcast_to(np.arange(n_cols), original.shape)
    roll_rows = np.mod(rows + trace_offset[None, :], n_rows)
    rolled = original[roll_rows, cols]

    # all zeros are treated as non-weighted (give non-zero fluxes)
    err = 0.1 * np.ones_like(rolled)
    mask = np.zeros_like(rolled).astype(bool)

    # unroll the trace using the Horne extract utility function for alignment:
    unrolled = _align_along_trace(rolled, n_rows // 2 - trace_offset)

    # ensure that mask is correctly unrolled back to its original alignment:
    np.testing.assert_allclose(unrolled, original)

    # Extract the spectrum from the non-flat image+trace
    extract_non_flat = HorneExtract(
        rolled, ArrayTrace(rolled, exact_trace), variance=err, mask=mask, unit=u.Jy
    )()

    # Also extract the spectrum from the image after alignment with a flat trace
    extract_flat = HorneExtract(
        unrolled, FlatTrace(unrolled, n_rows // 2), variance=err, mask=mask, unit=u.Jy
    )()

    # ensure both extractions are equivalent:
    assert_quantity_allclose(extract_non_flat.flux, extract_flat.flux)


def test_horne_bad_profile(mk_test_img):
    image = mk_test_img
    trace = FlatTrace(image, 3.0)
    extract = HorneExtract(
        image.data, trace, spatial_profile="bad_profile_name", variance=np.ones(image.data.shape)
    )
    with pytest.raises(ValueError, match="spatial_profile must be one of"):
        extract.spectrum

    extract = HorneExtract(
        image.data,
        trace,
        spatial_profile=models.Polynomial1D(2),
        variance=np.ones(image.data.shape),
    )
    with pytest.raises(ValueError, match="spatial_profile must be a"):
        extract.spectrum


def test_horne_nonfinite_column(mk_test_img):
    image = mk_test_img
    image.data[:, 4] = np.nan
    trace = FlatTrace(image, 3.0)
    extract = HorneExtract(
        image.data, trace, spatial_profile="gaussian", variance=np.ones(image.data.shape)
    )
    sp = extract.spectrum
    assert np.isnan(sp.flux.value[4])
    assert np.all(np.isfinite(sp.flux.value[:4]))
    assert np.all(np.isfinite(sp.flux.value[5:]))


def test_horne_no_bkgrnd(mk_test_img):
    # Test HorneExtract when using bkgrd_prof=None

    image = mk_test_img
    trace = FlatTrace(image, 3.0)
    extract = HorneExtract(image.data, trace, bkgrd_prof=None, variance=np.ones(image.data.shape))
    assert len(extract.spectrum.flux) == 10


def test_horne_interpolated_profile(mk_test_img):
    # basic test for HorneExtract using the spatial_profile == `interpolated_profile`
    # option add a perfectly gaussian source and make sure gaussian extraction
    # and self-profile extraction agree (since self=gaussian in this case)

    image = mk_test_img
    add_gaussian_source(image)  # add source across image, flat trace

    trace = FlatTrace(image, image.shape[0] // 2)

    # horne extraction using spatial_profile=='Gaussian'
    horne_extract_gauss = HorneExtract(
        image.data,
        trace,
        spatial_profile="gaussian",
        bkgrd_prof=None,
        variance=np.ones(image.data.shape),
    )

    # horne extraction with spatial_profile=='interpolated_profile'
    horne_extract_self = HorneExtract(
        image.data,
        trace,
        spatial_profile={"name": "interpolated_profile", "n_bins_interpolated_profile": 3},
        bkgrd_prof=None,
        variance=np.ones(image.data.shape),
    )

    assert_quantity_allclose(horne_extract_gauss.spectrum.flux, horne_extract_self.spectrum.flux)

    horne_extract_self = HorneExtract(
        image.data,
        trace,
        spatial_profile={"name": "interpolated_profile", "n_bins_interpolated_profile": 3},
        variance=np.ones(image.data.shape),
    )


def test_horne_interpolated_profile_norm(mk_test_img):
    # ensure that when using spatial_profile == `interpolated_profile`, the fit profile
    # represents the shape as a funciton of wavelength, and correctly accounts
    # for variations in trace position and flux.

    image = mk_test_img
    nrows, ncols = image.shape

    # create sawtooth pattern trace. right now, the _align_along_trace function
    # will rectify the trace to the integer-pixel level, so in this test
    # case the specturm will be totally straightened out.
    trace_shape = np.ones(ncols) * nrows // 2
    trace_shape[::2] += 4
    trace = ArrayTrace(image, trace_shape)

    # add gaussian source to image with increasing amplitude and that follows
    # along the trace. looks like a sawtooth pattern of increasing brightness
    add_gaussian_source(image, amps=np.linspace(1, 5, image.shape[1]), means=trace_shape)

    # horne extraction with spatial_profile=='interpolated_profile'
    # also tests that non-default parameters in the input dictionary format work
    ex = HorneExtract(
        image.data,
        trace,
        spatial_profile={
            "name": "interpolated_profile",
            "n_bins_interpolated_profile": 3,
            "interp_degree_interpolated_profile": (1, 1),
        },
        bkgrd_prof=None,
        variance=np.ones(image.data.shape),
    )

    # need to run produce extract.spectrum to access _interp_spatial_prof
    ex.spectrum

    # evaulate interpolated profile on entire grid
    interp_prof = ex._interp_spatial_prof(np.arange(ncols), np.arange(nrows)).T

    # the shifting position and amplitude should be accounted for, so the fit
    # spatial profile should just represent the shape as a function of
    # wavelength in this case, that is a gaussian with the area normalized at
    # each wavelength and a constant mean position

    # make sure that the fit spatial prof is normalized correctly
    assert_quantity_allclose(np.sum(interp_prof, axis=0), 1.0)

    # and that shifts in trace position are accounted for (to integer level)
    assert np.all(np.argmax(interp_prof, axis=0) == nrows // 2)


def test_horne_interpolated_nbins_fails(mk_test_img):
    # make sure that HorneProfile spatial_profile='interpolated_profile' correctly
    # fails when the number of samples is greater than the size of the image
    image = mk_test_img
    trace = FlatTrace(image, 5)

    with pytest.raises(ValueError):
        ex = HorneExtract(
            image.data,
            trace,
            spatial_profile={"name": "interpolated_profile", "n_bins_interpolated_profile": 100},
        )
        ex.spectrum


class TestMasksExtract:

    def mk_flat_gauss_img(self, nrows=200, ncols=160, nan_slices=None, add_noise=True):
        """
        Makes a flat gaussian image for testing, with optional added gaussian
        nosie and optional data values set to NaN. Variance is included, which
        is required by HorneExtract. Returns a Spectrum with flux, spectral
        axis, and uncertainty.
        """

        sigma_pix = 4
        col_model = models.Gaussian1D(amplitude=1, mean=nrows / 2, stddev=sigma_pix)
        spec2dvar = np.ones((nrows, ncols))
        noise = 0
        if add_noise:
            np.random.seed(7)
            sigma_noise = 1
            noise = np.random.normal(scale=sigma_noise, size=(nrows, ncols))

        index_arr = np.tile(np.arange(nrows), (ncols, 1))
        img = col_model(index_arr.T) + noise

        if nan_slices:  # add nans in data
            for s in nan_slices:
                img[s] = np.nan

        wave = np.arange(0, img.shape[1], 1)
        objectspec = Spectrum(
            spectral_axis=wave * u.m,
            flux=img * u.Jy,
            uncertainty=VarianceUncertainty(spec2dvar * u.Jy * u.Jy),
        )

        return objectspec

    def test_boxcar_fully_masked(self):
        """
        Test that the appropriate error is raised by `BoxcarExtract` when image
        is fully masked/NaN.
        """
        return

        img = self.mk_flat_gauss_img()
        trace = FitTrace(img)

        with pytest.raises(ValueError, match="Image is fully masked."):
            # fully NaN image
            img = np.zeros((4, 5)) * np.nan
            Background(img, traces=trace, width=2)

        with pytest.raises(ValueError, match="Image is fully masked."):
            # fully masked image (should be equivalent)
            img = NDData(np.ones((4, 5)), mask=np.ones((4, 5)))
            Background(img, traces=trace, width=2)

        # Now test that an image that isn't fully masked, but is fully masked
        # within the window determined by `width`, produces the correct result
        msg = "Image is fully masked within background window determined by `width`."
        with pytest.raises(ValueError, match=msg):
            img = self.mk_img(nrows=12, ncols=12, nan_slices=[np.s_[3:10, :]])
            Background(img, traces=FlatTrace(img, 6), width=7)



"""Utility functions to parse main NIST table."""

import numpy as np
from astropy.table import Table, vstack

__all__ = []


def sort_table_by_element(table, elem_list):
    """Build table based on list of elements

    Parameters
    ----------
    table: astropy table
        Table to sort
    elem_list: list
        list of strings to sort table by

    Returns
    -------
    element_filtered_table: astropytable
        Filtered table based on inputs
    """

    filtered_table_list = [table[np.where(table['Element'] == elem)] for elem in elem_list]
    element_filtered_table = vstack(filtered_table_list)

    return element_filtered_table


def sort_table_by_wavelength(table, min_wave, max_wave):
    """Build table off of wavelength ranges

    Parameters
    ----------
    min_wave: float
        Lower bound wavelength to filter on
    max_wave: float
        Upper bound wavelength to filter on

    Returns
    -------
    wave_filtered_table: astropytable
        Filtered table based on inputs
    """

    assert min_wave < max_wave, "Minimum wavelength greater than maximum wavelength."
    wave_filtered_table = table[
        np.where(
            (table['Wavelength'] >= min_wave) & (table['Wavelength'] <= max_wave)
        )
    ]

    return wave_filtered_table


def main():
    """A little example.
    """
    t = Table.read('data/line_lists/NIST/NIST_combined.csv', format='csv')
    elements = ['He I', 'Ne I', 'Ar I']
    sorted_by_elem = sort_table_by_element(t, elements)
    sorted_by_wave = sort_table_by_wavelength(t, 2000, 3000)

    print(sorted_by_wave)
    print(sorted_by_elem)


if __name__ == "__main__":
    main()



import specutils
from astropy.utils import minversion

__all__ = ["Spectrum"]

SPECUTILS_LT_2 = not minversion(specutils, "2.0.dev")

if SPECUTILS_LT_2:
    from specutils import Spectrum1D as Spectrum
else:
    from specutils import Spectrum  # noqa: F401



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Literal

import numpy as np
from astropy.modeling import Model, fitting, models
from astropy.nddata import NDData
from astropy.stats import gaussian_sigma_to_fwhm
from astropy.utils.decorators import deprecated

from specreduce.core import _ImageParser

__all__ = ["Trace", "FlatTrace", "ArrayTrace", "FitTrace"]


@dataclass
class Trace:
    """
    Basic tracing class that by default traces the middle of the image.

    Parameters
    ----------
    image : `~astropy.nddata.NDData`-like or array-like, required
        Image to be traced

    Attributes
    ----------
    shape : tuple
        Shape of the array describing the trace
    """

    image: NDData

    def __post_init__(self):
        self.trace_pos = self.image.shape[0] / 2
        self.trace = np.ones_like(self.image[0]) * self.trace_pos

        # masking options not relevant for basic Trace
        self._mask_treatment = None
        self._valid_mask_treatment_methods = [None]

        # eventually move this to common SpecreduceOperation base class
        self.validate_masking_options()

    def __getitem__(self, i):
        return self.trace[i]

    @property
    def shape(self):
        return self.trace.shape

    def validate_masking_options(self):
        if self.mask_treatment not in self.valid_mask_treatment_methods:
            raise ValueError(
                f"`mask_treatment` {self.mask_treatment} not one "
                f"of {self.valid_mask_treatment_methods}"
            )  # noqa

    def shift(self, delta):
        """
        Shift the trace by delta pixels perpendicular to the axis being traced

        Parameters
        ----------
        delta : float
            Shift to be applied to the trace
        """
        # act on self.trace.data to ignore the mask and then re-mask when calling _bound_trace
        self.trace = np.asarray(self.trace.data) + delta
        self._bound_trace()

    def _bound_trace(self):
        """
        Mask trace positions that are outside the upper/lower bounds of the image.
        """
        ny = self.image.shape[0]
        self.trace = np.ma.masked_outside(self.trace, 0, ny - 1)

    def __add__(self, delta):
        """
        Return a copy of the trace shifted "forward" by delta pixels perpendicular to the axis
        being traced
        """
        copy = deepcopy(self)
        copy.shift(delta)
        return copy

    def __sub__(self, delta):
        """
        Return a copy of the trace shifted "backward" by delta pixels perpendicular to the axis
        being traced
        """
        return self.__add__(-delta)

    @property
    def mask_treatment(self):
        return self._mask_treatment

    @property
    def valid_mask_treatment_methods(self):
        return self._valid_mask_treatment_methods


@dataclass
class FlatTrace(Trace, _ImageParser):
    """
    Trace that is constant along the axis being traced.

    Example: ::

        trace = FlatTrace(image, trace_pos)

    Parameters
    ----------
    trace_pos : float
        Position of the trace
    """

    trace_pos: float

    def __post_init__(self):
        self.image = self._parse_image(self.image)

        self.set_position(self.trace_pos)

        # masking options not relevant for basic Trace
        self._mask_treatment = None
        self._valid_mask_treatment_methods = [None]

    def set_position(self, trace_pos):
        """
        Set the trace position within the image

        Parameters
        ----------
        trace_pos : float
            Position of the trace
        """
        if trace_pos < 1:
            raise ValueError("`trace_pos` must be positive.")
        self.trace_pos = trace_pos
        self.trace = np.ones_like(self.image.data[0]) * self.trace_pos
        self._bound_trace()


@dataclass
class ArrayTrace(Trace, _ImageParser):
    """
    Define a trace given an array of trace positions.

    Parameters
    ----------
    trace : `numpy.ndarray` or `numpy.ma.MaskedArray`
        Array containing trace positions.
    """

    trace: np.ndarray

    def __post_init__(self):

        # masking options not relevant for ArrayTrace. any non-finite or masked
        # data in `image` will not affect array trace
        self._mask_treatment = None
        self._valid_mask_treatment_methods = [None]

        # masked array will have a .data, regular array will not.
        trace_data = getattr(self.trace, "data", self.trace)

        # but we do need to mask uncaught non-finite values in input trace array
        # which should also be combined with any existing mask in the input `trace`
        if hasattr(self.trace, "mask"):
            total_mask = np.logical_or(self.trace.mask, ~np.isfinite(trace_data))
        else:
            total_mask = ~np.isfinite(trace_data)

        # always work with masked array, even if there is no masked
        # or nonfinite data, in case padding is needed. if not, mask will be
        # dropped at the end and a regular array will be returned.
        self.trace = np.ma.MaskedArray(trace_data, total_mask)

        self.image = self._parse_image(self.image)

        nx = self.image.shape[1]
        nt = len(self.trace)
        if nt != nx:
            if nt > nx:
                # truncate trace to fit image
                self.trace = self.trace[0:nx]
            else:
                # assume trace starts at beginning of image and pad out trace to fit.
                # padding will be the last value of the trace, but will be masked out.
                padding = np.ma.MaskedArray(np.ones(nx - nt) * self.trace[-1], mask=True)
                self.trace = np.ma.hstack([self.trace, padding])

        self._bound_trace()

        # warn if entire trace is masked
        if np.all(self.trace.mask):
            warnings.warn("Entire trace array is masked.")

        # and return plain array if nothing is masked
        if not np.any(self.trace.mask):
            self.trace = self.trace.data


@dataclass
class FitTrace(Trace, _ImageParser):
    """
    Trace the spectrum aperture in an image.

    Bins along the image's dispersion (wavelength) direction, finds each
    bin's peak cross-dispersion (spatial) pixel, and uses a model to
    interpolate the function fitted to the peaks as a final trace. The
    number of bins, peak finding algorithm, and model used for fitting
    are customizable by the user.


    Example: ::

        trace = FitTrace(image, peak_method='gaussian', guess=trace_pos)

    Parameters
    ----------
    image : `~astropy.nddata.NDData`-like or array-like, required
        The image over which to run the trace. Assumes cross-dispersion
        (spatial) direction is axis 0 and dispersion (wavelength)
        direction is axis 1.
    bins : int, optional
        The number of bins in the dispersion (wavelength) direction
        into which to divide the image. If not set, defaults to one bin
        per dispersion (wavelength) pixel in the given image. If set,
        requires at least 4 or N bins for a degree N ``trace_model``,
        whichever is greater. [default: ``None``]
    guess : int, optional
        A guess at the trace's location in the cross-dispersion
        (spatial) direction. If set, overrides the normal max peak
        finder. Good for tracing a fainter source if multiple traces
        are present. [default: ``None``]
    window : int, optional
        Fit the trace to a region with size ``window * 2`` around the
        guess position. Useful for tracing faint sources if multiple
        traces are present, but potentially bad if the trace is
        substantially bent or warped. [default: ``None``]
    disp_bounds
        The lower and upper bounds of the pixel range along the dispersion
        axis that is used when fitting the trace model. If ``None``,
        defaults to the entire range of pixels along the dispersion axis.
        [default: ``None``]
    trace_model : one of `~astropy.modeling.polynomial.Chebyshev1D`,\
            `~astropy.modeling.polynomial.Legendre1D`,\
            `~astropy.modeling.polynomial.Polynomial1D`,\
            or `~astropy.modeling.spline.Spline1D`, optional
        The 1-D polynomial model used to fit the trace to the bins' peak
        pixels. Spline1D models are fit with Astropy's
        'SplineSmoothingFitter', generic linear models are fit with the
        'LinearLSQFitter', while the other models are fit with the
        'LMLSQFitter'. [default: ``models.Polynomial1D(degree=1)``]
    peak_method : string, optional
        One of ``gaussian``, ``centroid``, or ``max``.
        ``gaussian``: Fits a gaussian to the window within each bin and
        adopts the central value as the peak. May work best with fewer
        bins on faint targets. (Based on the "kosmos" algorithm from
        James Davenport's same-named repository.)
        ``centroid``: Takes the centroid of the window within in bin.
        ``max``: Saves the position with the maximum flux in each bin.
        [default: ``max``]
    mask_treatment
        Specifies how to handle masked or non-finite values in the input image.
        The fit cannot handle non-finite values, so only the ``apply``, ``propagate``,
        ``apply_nan_only`` options are supported. The ``apply`` option combines
        the existing mask with the mask derived from non-finite values, ``propagate``
        expands the mask along the cross-dispersion axis (that is, a masked pixel
        results in the whole cross-dispersion slice being masked), and
        ``apply_nan_only`` drops the existing mask and replaces it with a mask
        derived from non-finite values.
    """

    bins: int | None = None
    guess: float | None = None
    window: int | None = None
    disp_bounds: tuple[float, float] | None = None
    trace_model: Model = field(default=models.Polynomial1D(degree=1))
    peak_method: Literal["gaussian", "centroid", "max"] = "max"
    _crossdisp_axis: int = 0
    _disp_axis: int = 1
    mask_treatment: Literal["apply", "propagate", "apply_nan_only"] = "apply"
    _valid_mask_treatment_methods = ("apply", "propagate", "apply_nan_only")
    # for testing purposes only, save bin peaks if requested
    _save_bin_peaks_testing: bool = True

    def __post_init__(self):

        # Parse image, including masked/nonfinite data handling based on
        # choice of `mask_treatment`. returns a Spectrum1D
        if self.mask_treatment not in self._valid_mask_treatment_methods:
            raise ValueError(
                "`mask_treatment` must be one of " f"{self._valid_mask_treatment_methods}"
            )

        self.image = self._parse_image(
            self.image, disp_axis=self._disp_axis, mask_treatment=self.mask_treatment
        )

        # _parse_image returns a Spectrum1D. convert this to a masked array
        # for ease of calculations here (even if there is no masked data).
        # Note: uncertainties are dropped, this should also be addressed at
        # some point probably across the package.
        img = np.ma.masked_array(self.image.data, self.image.mask)
        self._mask_temp = self.image.mask

        # validate input arguments
        valid_peak_methods = ("gaussian", "centroid", "max")
        if self.peak_method not in valid_peak_methods:
            raise ValueError(f"peak_method must be one of {valid_peak_methods}")

        if self._crossdisp_axis != 0:
            raise ValueError("cross-dispersion axis must equal 0")

        if self._disp_axis != 1:
            raise ValueError("dispersion axis must equal 1")

        valid_models = (models.Spline1D, models.Legendre1D, models.Chebyshev1D, models.Polynomial1D)
        if not isinstance(self.trace_model, valid_models):
            raise ValueError(
                "trace_model must be one of " f"{', '.join([m.name for m in valid_models])}."
            )

        self.dlim = self.disp_bounds if self.disp_bounds is not None else (0, img.shape[
            self._disp_axis])
        cols = self.dlim[1]
        model_deg = self.trace_model.degree
        if self.bins is None:
            self.bins = cols
        elif self.bins < 4:
            # many of the Astropy model fitters require four points at minimum
            raise ValueError("bins must be >= 4")
        elif self.bins <= model_deg:
            raise ValueError(f"bins must be > {model_deg} for " f"a degree {model_deg} model.")
        elif self.bins > cols:
            raise ValueError(
                f"bins must be <= {cols}, the length of the " "image's spatial direction"
            )

        if not isinstance(self.bins, int):
            warnings.warn("TRACE: Converting bins to int")
            self.bins = int(self.bins)

        if self.window is not None and (
            self.window > img.shape[self._disp_axis] or self.window < 1
        ):
            raise ValueError(
                f"window must be >= 2 and less than {cols}, the "
                "length of the image's spatial direction"
            )
        elif self.window is not None and not isinstance(self.window, int):
            warnings.warn("TRACE: Converting window to int")
            self.window = int(self.window)

        # fit the trace
        self._fit_trace(img)

    def _fit_trace(self, img):

        yy = np.arange(img.shape[self._crossdisp_axis])

        # set max peak location by user choice or wavelength with max avg flux
        ztot = img.mean(axis=self._disp_axis)
        peak_y = self.guess if self.guess is not None else ztot.argmax()
        # NOTE: peak finder can be bad if multiple objects are on slit

        if self.peak_method == "gaussian":

            # guess the peak width as the FWHM, roughly converted to gaussian sigma
            yy_above_half_max = np.sum(ztot > (ztot.max() / 2))
            width_guess = yy_above_half_max / gaussian_sigma_to_fwhm

            # enforce some (maybe sensible?) rules about trace peak width
            width_guess = 2 if width_guess < 2 else 25 if width_guess > 25 else width_guess

            # fit a Gaussian to peak for fall-back answer, but don't use yet
            g1d_init = models.Gaussian1D(amplitude=ztot.max(), mean=peak_y, stddev=width_guess)
            offset_init = models.Const1D(np.ma.median(ztot))
            profile = g1d_init + offset_init

            fitter = fitting.DogBoxLSQFitter()
            popt_tot = fitter(profile, yy, ztot)

        # restrict fit to window (if one exists)
        ilum2 = (
            yy
            if self.window is None
            else yy[np.arange(peak_y - self.window, peak_y + self.window, dtype=int)]
        )

        # check if everything in window region is masked
        if img[ilum2].mask.all():
            raise ValueError(
                "All pixels in window region are masked. Check "
                "for invalid values or use a larger window value."
            )

        x_bins = np.linspace(self.dlim[0], self.dlim[1], self.bins + 1, dtype=int)
        y_bins = np.tile(np.nan, self.bins)

        warn_bins = []
        for i in range(self.bins):

            # binned columns, averaged along disp. axis.
            # or just a single, unbinned column if no bins
            z_i = img[ilum2, x_bins[i] : x_bins[i + 1]].mean(axis=self._disp_axis)

            # if this bin is fully masked, set bin peak to NaN so it can be
            # filtered in the final fit to all bin peaks for the trace
            if z_i.mask.all() or (~np.isfinite(z_i)).all():
                warn_bins.append(i)
                y_bins[i] = np.nan
                continue

            if self.peak_method == "gaussian":

                # if bin is fully 0, set bin peak to nan so it doesn't bias the
                # all-bin fit. DogBoxLSQFitter, which is always used for the bin
                # center fits when peak_method is gaussian, does not like all zeros.
                if np.all(z_i == 0.0):
                    y_bins[i] = np.nan
                    continue

                peak_y_i = ilum2[z_i.argmax()]

                yy_i_above_half_max = np.sum(z_i > (z_i.max() / 2))
                width_guess_i = yy_i_above_half_max / gaussian_sigma_to_fwhm

                # NOTE: original KOSMOS code mandated width be greater than 2
                # (to avoid cosmic rays) and less than 25 (to avoid fitting noise).
                # we should extract values from img to derive similar limits
                # width_guess_i = (2 if width_guess_i < 2
                #                  else 25 if width_guess_i > 25
                #                  else width_guess_i)

                g1d_init_i = models.Gaussian1D(
                    amplitude=z_i.max(), mean=peak_y_i, stddev=width_guess_i
                )
                offset_init_i = models.Const1D(np.ma.median(z_i))

                profile_i = g1d_init_i + offset_init_i
                popt_i = fitter(profile_i, ilum2[~z_i.mask], z_i.data[~z_i.mask])

                # if gaussian fits off chip, then fall back to previous answer
                if not ilum2.min() <= popt_i.mean_0 <= ilum2.max():
                    y_bins[i] = popt_tot.mean_0.value
                else:
                    y_bins[i] = popt_i.mean_0.value
                    popt_tot = popt_i

            elif self.peak_method == "centroid":
                z_i_cumsum = np.cumsum(z_i)
                # find the interpolated index where the cumulative array reaches
                # half the total cumulative values
                y_bins[i] = np.interp(z_i_cumsum[-1] / 2.0, z_i_cumsum, ilum2)

                # NOTE this reflects current behavior, should eventually be changed
                # to set to nan by default (or zero fill / interpoate option once
                # available)

            elif self.peak_method == "max":
                # TODO: implement smoothing with provided width
                y_bins[i] = ilum2[z_i.argmax()]

                # NOTE: a fully masked should eventually be changed to set to
                # nan by default (or zero fill / interpoate option once available)

        # warn about fully-masked bins
        if len(warn_bins) > 0:

            # if there are a ton of bins, we don't want to print them all out
            if len(warn_bins) > 20:
                warn_bins = warn_bins[0:10] + ["..."] + [warn_bins[-1]]

            warnings.warn(
                f"All pixels in {'bins' if len(warn_bins) else 'bin'} "
                f"{', '.join([str(x) for x in warn_bins])}"
                " are fully masked. Setting bin"
                f" peak{'s' if len(warn_bins) else ''} to NaN."
            )

        # recenter bin positions
        x_bins = (x_bins[:-1] + x_bins[1:]) / 2

        # interpolate the fitted trace over the entire wavelength axis

        # for testing purposes only, save bin peaks if requested
        if self._save_bin_peaks_testing:
            self._bin_peaks = (x_bins, y_bins)

        # filter non-finite bin peaks before filtering to all bin peaks
        y_finite = np.where(np.isfinite(y_bins))[0]
        if y_finite.size > 0:
            x_bins = x_bins[y_finite]
            y_bins = y_bins[y_finite]

            # use given model to bin y-values; interpolate over all wavelengths
            if isinstance(self.trace_model, models.Spline1D):
                fitter = fitting.SplineSmoothingFitter()
            elif self.trace_model.linear:
                fitter = fitting.LinearLSQFitter()
            else:
                fitter = fitting.LMLSQFitter()

            self._y_bins = y_bins
            self.trace_model_fit = fitter(self.trace_model, x_bins, y_bins)

            trace_x = np.arange(img.shape[self._disp_axis])
            trace_y = self.trace_model_fit(trace_x)
        else:
            warnings.warn("TRACE ERROR: No valid points found in trace")
            trace_y = np.tile(np.nan, img.shape[self._disp_axis])

        self.trace = np.ma.masked_invalid(trace_y)


@deprecated("1.3", alternative="FitTrace")
@dataclass
class KosmosTrace(FitTrace):
    """
    This class is pending deprecation. Please use `FitTrace` instead.
    """

    __doc__ += FitTrace.__doc__
    pass



import pytest

from specreduce.core import SpecreduceOperation


def test_sro_call():
    sro = SpecreduceOperation()
    with pytest.raises(NotImplementedError):
        sro()


def test_sro_as_function():

    class TestSRO(SpecreduceOperation):
        def __call__(self, x):
            return x**2

    assert TestSRO.as_function(6) == 36

    class TestSRO(SpecreduceOperation):
        def __call__(self, *nargs):
            return 5

    with pytest.raises(NotImplementedError, match="There is not a way"):
        TestSRO.as_function(6) == 36



import numpy as np
from astropy import units as u

from specreduce.compat import Spectrum
from specreduce.core import _ImageParser
from specreduce.extract import HorneExtract
from specreduce.tracing import FlatTrace


# (for use inside tests)
def compare_images(all_images, key, collection, compare='s1d'):
    # save default values used for spectral axis and uncertainty when they are not
    # available from the image object or provided by the user
    unc_def = np.ones_like(all_images['arr'])
    sax_def = np.arange(unc_def.shape[1]) * u.pix

    # was input converted to Spectrum?
    assert isinstance(collection[key], Spectrum), (f"image '{key}' not "
                                                   "of type Spectrum")

    # do key's fluxes match its comparison's fluxes?
    assert np.allclose(collection[key].data,
                       collection[compare].data), (f"images '{key}' and "
                                                   f"'{compare}' have unequal "
                                                   "flux values")

    # if the image came with a spectral axis, was it kept? if not, was the
    # default spectral axis in pixels applied?
    sax_provided = hasattr(all_images[key], 'spectral_axis')
    assert np.allclose(collection[key].spectral_axis,
                       (all_images[key].spectral_axis if sax_provided
                        else sax_def)), (f"spectral axis of image '{key}' does "
                                         f"not match {'input' if sax_provided else 'default'}")

    # if the image came with an uncertainty, was it kept? if not, was the
    # default uncertainty created?
    unc_provided = hasattr(all_images[key], 'uncertainty')
    assert np.allclose(collection[key].uncertainty.array,
                       (all_images[key].uncertainty.array if unc_provided
                        else unc_def)), (f"uncertainty of image '{key}' does "
                                         f"not match {'input' if unc_provided else 'default'}")

    # were masks created despite none being given? (all indices should be False)
    assert (getattr(collection[key], 'mask', None)
            is not None), f"no mask was created for image '{key}'"
    assert np.all(collection[key].mask == 0), ("mask not all False "
                                               f"for image '{key}'")


# test consistency of general image parser results
def test_parse_general(all_images):

    all_images_parsed = {k: _ImageParser()._parse_image(im)
                         for k, im in all_images.items()}
    for key in all_images_parsed.keys():
        compare_images(all_images, key, all_images_parsed)


# use verified general image parser results to check HorneExtract's image parser
def test_parse_horne(all_images):
    # save default values used for uncertainty when it is
    # available from the image object or provided by the user
    unc_def = np.ones_like(all_images['arr'])

    # HorneExtract's parser is more stringent than the general one, hence the
    # separate test. Given proper inputs, both should produce the same results.
    images_collection = {k: {} for k in all_images.keys()}

    for key, col in images_collection.items():
        img = all_images[key]
        col['general'] = _ImageParser()._parse_image(img)

        if hasattr(all_images[key], 'uncertainty'):
            defaults = {}
        else:
            # save default values of attributes used in general parser when
            # they are not available from the image object. HorneExtract always
            # requires a variance, so it's chosen here to be on equal footing
            # with the general case
            defaults = {'variance': unc_def,
                        'mask': ~np.isfinite(img),
                        'unit': getattr(img, 'unit', u.DN)}

        col[key] = HorneExtract(img, FlatTrace(img, 2))._parse_image(img, **defaults)

        compare_images(all_images, key, col, compare='general')



"""
Utilities for defining, loading, and handling spectroscopic calibration data
"""

import warnings
from pathlib import Path
from typing import Sequence, Literal
from urllib.error import URLError

from astropy import units as u
from astropy.table import Table, vstack, QTable
from astropy.utils.data import download_file
from astropy.utils.exceptions import AstropyUserWarning
from astropy.coordinates import SpectralCoord
from specutils.utils.wcs_utils import vac_to_air

from specreduce.compat import Spectrum

__all__ = [
    'get_available_line_catalogs',
    'load_pypeit_calibration_lines',
    'load_MAST_calspec',
    'load_onedstds',
    'AtmosphericExtinction',
    'AtmosphericTransmission'
]

SUPPORTED_EXTINCTION_MODELS = [
    "kpno",
    "ctio",
    "apo",
    "lapalma",
    "mko",
    "mtham",
    "paranal"
]

SPECPHOT_DATASETS = [
    "bstdscal",
    "ctiocal",
    "ctionewcal",
    "eso",
    "gemini",
    "iidscal",
    "irscal",
    "oke1990",
    "redcal",
    "snfactory",
    "spec16cal",
    "spec50cal",
    "spechayescal"
]

PYPEIT_CALIBRATION_LINELISTS = [
    'Ne_IR_MOSFIRE',
    'ArII',
    'CdI',
    'OH_MOSFIRE_H',
    'OH_triplespec',
    'Ar_IR_MOSFIRE',
    'OH_GNIRS',
    'ThAr_XSHOOTER_VIS',
    'ThAr_MagE',
    'HgI',
    'NeI',
    'XeI',
    'OH_MODS',
    'ZnI',
    'OH_GMOS',
    'CuI',
    'ThAr_XSHOOTER_VIS_air',
    'ThAr_XSHOOTER_UVB',
    'OH_NIRES',
    'HeI',
    'FeI',
    'OH_MOSFIRE_J',
    'KrI',
    'Cd_DeVeny1200',
    'Ar_IR_GNIRS',
    'OH_MOSFIRE_Y',
    'ThAr',
    'FeII',
    'OH_XSHOOTER',
    'OH_FIRE_Echelle',
    'OH_MOSFIRE_K',
    'OH_R24000',
    'Hg_DeVeny1200',
    'ArI'
]

SPECREDUCE_DATA_URL = ("https://raw.githubusercontent.com/astropy/specreduce-data/"
                       "main/specreduce_data/reference_data/")

PYPEIT_DATA_URL = ("https://raw.githubusercontent.com/pypeit/"
                   "pypeit/release/pypeit/data/")


def get_available_line_catalogs() -> dict:
    """
    Returns a dictionary of available line catalogs. Currently only ``pypeit``
    catalogs are fully supported.
    """
    return {
        'pypeit': PYPEIT_CALIBRATION_LINELISTS
    }


def load_pypeit_calibration_lines(
        lamps: Sequence | None = None,
        wave_air: bool = False,
        cache: bool | Literal['update'] = True,
        show_progress: bool = False
) -> QTable | None:
    """
    Load reference calibration lines from ``pypeit`` linelists. The ``pypeit`` linelists are
    well-curated and have been tested across a wide range of spectrographs. The available
    linelists are defined by ``PYPEIT_CALIBRATION_LINELISTS``.

    Parameters
    ----------
    lamps : Lamp string, comma-separated list of lamps, or sequence of lamps to include in
        output reference linelist. The parlance of "lamp" is retained here for consistency
        with its use in ``pypeit`` and elsewhere. In several of the supported cases the
        "lamp" is the sky itself (e.g. OH lines in the near-IR).
        The available lamps are defined by ``PYPEIT_CALIBRATION_LINELISTS``.

    wave_air : If True, convert the vacuum wavelengths used by ``pypeit`` to air wavelengths.

    cache : Toggle caching of downloaded data

    show_progress : Show download progress bar

    Returns
    -------
    linelist:
        Table containing the combined calibration line list. ``pypeit`` linelists have the
        following columns:
        * ``ion``: Ion or molecule generating the line.
        * ``wavelength``: Vacuum wavelength of the line in Angstroms.
        * ``NIST``: Flag denoting if NIST is the ultimate reference for the line's wavelength.
        * ``Instr``: ``pypeit``-specific instrument flag.
        * ``amplitude``: Amplitude of the line. Beware, not consistent between lists.
        * ``Source``: Source of the line information.
    """
    if lamps is None:
        return None

    if not isinstance(lamps, Sequence):
        raise ValueError(f"Invalid calibration lamps specification: {lamps}")

    if isinstance(lamps, str):
        if ',' in lamps:
            lamps = [lamp.strip() for lamp in lamps.split(',')]
        else:
            lamps = [lamps]

    linelists = []
    for lamp in lamps:
        if lamp in PYPEIT_CALIBRATION_LINELISTS:
            data_url = f"{PYPEIT_DATA_URL}/arc_lines/lists/{lamp}_lines.dat"
            try:
                data_path = download_file(data_url, cache=cache,
                                          show_progress=show_progress,
                                          pkgname='specreduce')
                linelists.append(Table.read(data_path, format='ascii.fixed_width', comment='#'))
            except URLError as e:
                warnings.warn(f"Downloading of {data_url} failed: {e}", AstropyUserWarning)
        else:
            warnings.warn(
                f"{lamp} not in the list of supported calibration "
                f"line lists: {PYPEIT_CALIBRATION_LINELISTS}."
            )
    if len(linelists) == 0:
        warnings.warn(f"No calibration lines loaded from {lamps}.")
        linelist = None
    else:
        linelist = QTable(vstack(linelists))
        linelist.rename_column('wave', 'wavelength')
        # pypeit linelists use vacuum wavelengths in angstroms
        linelist['wavelength'] *= u.Angstrom
        if wave_air:
            linelist['wavelength'] = vac_to_air(linelist['wavelength'])
    return linelist


def load_MAST_calspec(
        filename: str | Path,
        cache: bool | Literal['update'] = True,
        show_progress: bool = False
) -> Spectrum | None:
    """
    Load a standard star spectrum from the ``calspec`` database at MAST. These spectra are provided
    in FITS format and are described in detail at:

    https://www.stsci.edu/hst/instrumentation/reference-data-for-calibration-and-tools/astronomical-catalogs/calspec

    If ``remote`` is True, the spectrum will be downloaded from MAST. Set ``remote`` to False to
    load a local file.

    .. note:: This function requires ``synphot`` to be installed separately.

    Parameters
    ----------
    filename : FITS filename of a standard star spectrum, e.g. g191b2b_005.fits.
        If this is a local file, it will be loaded. If not, then a download from
        MAST will be attempted.

    cache : Toggle whether downloaded data is cached or not.

    show_progress : Toggle whether download progress bar is shown.

    Returns
    -------
    spectrum : If the spectrum can be loaded, return it as a `~specutils.Spectrum`.
        Otherwise return None. The spectral_axis units are Å and the flux units are milli-Janskys.
    """
    filename = Path(filename)
    if filename.exists() and filename.is_file():
        file_path = filename
    else:
        try:
            data_url = f"https://archive.stsci.edu/hlsps/reference-atlases/cdbs/calspec/{filename}"
            file_path = download_file(data_url, cache=cache,
                                      show_progress=show_progress,
                                      pkgname='specreduce')
        except URLError as e:
            warnings.warn(f"Downloading of {filename} failed: {e}", AstropyUserWarning)
            file_path = None

    if file_path is None:
        return None
    else:
        import synphot
        _, wave, flux = synphot.specio.read_fits_spec(file_path)

        # DEV: pllim does not think this is necessary at all but whatever.
        # the calspec data stores flux in synphot's FLAM units. convert to flux units
        # supported directly by astropy.units. mJy is chosen since it's the JWST
        # standard and can easily be converted to/from AB magnitudes.
        flux_mjy = synphot.units.convert_flux(wave, flux, u.mJy)
        spectrum = Spectrum(spectral_axis=wave, flux=flux_mjy)
        return spectrum


def load_onedstds(
        dataset: str = "snfactory",
        specfile: str = "EG131.dat",
        cache: bool | Literal['update'] = True,
        show_progress: bool = False
) -> Spectrum | None:
    """
    This is a convenience function for loading a standard star spectrum from the 'onedstds'
    dataset in the ``specreduce_data`` package. They will be downloaded from the
    repository on GitHub and cached by default.

    Parameters
    ----------
    dataset : Standard star spectrum database. Valid options are described
        in :ref:`specphot_standards`.

    specfile : Filename of the standard star spectrum.

    cache : Enable caching of downloaded data.

    show_progress : Show download progress bar if data is downloaded.

    Returns
    -------
    spectrum : If the spectrum can be loaded, return it as a `~specutils.Spectrum`.
        Otherwise return None. The spectral_axis units are Å and the flux units are milli-Janskys.
    """
    if dataset not in SPECPHOT_DATASETS:
        msg = (f"Specfied dataset, {dataset}, not in list of supported datasets of "
               f"spectrophotometric standard stars: f{SPECPHOT_DATASETS}")
        warnings.warn(msg, AstropyUserWarning)
        return None

    try:
        data_path = download_file(f"{SPECREDUCE_DATA_URL}/onedstds/{dataset}/{specfile}",
                                  cache=cache, show_progress=show_progress, pkgname="specreduce")
        t = Table.read(data_path, format="ascii", names=['wavelength', 'ABmag', 'binsize'])
    except URLError as e:
        msg = f"Can't load {specfile} from {dataset}: {e}."
        warnings.warn(msg, AstropyUserWarning)
        return None

    # the specreduce_data standard star spectra all provide wavelengths in angstroms
    spectral_axis = t['wavelength'].data * u.angstrom

    # the specreduce_data standard star spectra all provide fluxes in AB mag
    flux = t['ABmag'].data * u.ABmag
    flux = flux.to(u.mJy)  # convert to linear flux units
    spectrum = Spectrum(spectral_axis=spectral_axis, flux=flux)
    return spectrum


class AtmosphericExtinction(Spectrum):
    """
    Spectrum container for atmospheric extinction in magnitudes as a function of wavelength.
    If extinction and spectral_axis are provided, this will use them to build a custom model.
    If they are not, the 'model' parameter will be used to lookup and load a pre-defined
    atmospheric extinction model from the ``specreduce_data`` package.

    Parameters
    ----------
    model : str
        Name of atmospheric extinction model provided by ``specreduce_data``. Valid
        options are:

        * kpno - Kitt Peak National Observatory (default)
        * ctio - Cerro Tololo International Observatory
        * apo - Apache Point Observatory
        * lapalma - Roque de los Muchachos Observatory, La Palma, Canary Islands
        * mko - Mauna Kea Observatories
        * mtham - Lick Observatory, Mt. Hamilton station
        * paranal - European Southern Observatory, Cerro Paranal station

    extinction : float, `~astropy.units.Quantity`, or `None`, optional
        Provides extinction data for this spectrum. Used along with
        spectral_axis to build custom atmospheric extinction model. If no units are provided,
        assumed to be given in magnitudes.

    spectral_axis : `~astropy.coordinates.SpectralCoord`, `~astropy.units.Quantity`, or `None`, optional
        Dispersion information with the same shape as the last (or only)
        dimension of flux, or one greater than the last dimension of flux
        if specifying bin edges. Used along with flux to build custom atmospheric
        extinction model.

    Attributes
    ----------
    extinction_mag : `~astropy.units.Quantity`
        Extinction expressed in dimensionless magnitudes

    transmission : `~astropy.units.Quantity`
        Extinction expressed as fractional transmission

    """  # noqa: E501
    def __init__(
        self,
        model: str = "kpno",
        extinction: Sequence[float] | u.Quantity | None = None,
        spectral_axis: SpectralCoord | u.Quantity | None = None,
        cache: bool | Literal['update'] = True,
        show_progress: bool = False,
        **kwargs: str
    ) -> None:
        if extinction is not None:
            if not isinstance(extinction, u.Quantity):
                warnings.warn(
                    "Input extinction is not a Quanitity. Assuming it is given in magnitudes...",
                    AstropyUserWarning
                )
                extinction = u.Magnitude(
                    extinction,
                    u.MagUnit(u.dimensionless_unscaled)
                ).to(u.dimensionless_unscaled)  # Spectrum wants this to be linear
            elif isinstance(extinction, (u.LogUnit, u.Magnitude)) or extinction.unit == u.mag:
                # if in log or magnitudes, recast into Magnitude with dimensionless physical units
                extinction = u.Magnitude(
                    extinction.value,
                    u.MagUnit(u.dimensionless_unscaled)
                ).to(u.dimensionless_unscaled)
            elif extinction.unit != u.dimensionless_unscaled:
                # if we're given something linear that's not dimensionless_unscaled,
                # it's an error
                msg = "Input extinction must have unscaled dimensionless units."
                raise ValueError(msg)

        if extinction is None and spectral_axis is None:
            if model not in SUPPORTED_EXTINCTION_MODELS:
                msg = (
                    f"Requested extinction model, {model}, not in list "
                    f"of available models: {SUPPORTED_EXTINCTION_MODELS}"
                )
                raise ValueError(msg)

            data_file = download_file(f"{SPECREDUCE_DATA_URL}/extinction/{model}extinct.dat",
                                      cache=cache, show_progress=show_progress,
                                      pkgname='specreduce')
            t = Table.read(data_file, format="ascii", names=['wavelength', 'extinction'])

            # the specreduce_data models all provide wavelengths in angstroms
            spectral_axis = t['wavelength'].data * u.angstrom

            # the specreduce_data models all provide extinction in magnitudes at an airmass of 1
            extinction = u.Magnitude(
                t['extinction'].data,
                u.MagUnit(u.dimensionless_unscaled)
            ).to(u.dimensionless_unscaled)

        if spectral_axis is None:
            msg = "Missing spectral axis for input extinction data."
            raise ValueError(msg)

        super(AtmosphericExtinction, self).__init__(
            flux=extinction,
            spectral_axis=spectral_axis,
            unit=u.dimensionless_unscaled,
            **kwargs
        )

    @property
    def extinction_mag(self) -> u.Quantity:
        """
        This property returns the extinction in magnitudes
        """
        return self.flux.to(u.mag(u.dimensionless_unscaled))

    @property
    def transmission(self) -> u.Quantity:
        """
        This property returns the transmission as a fraction between 0 and 1
        """
        return self.flux


class AtmosphericTransmission(AtmosphericExtinction):
    """
    Spectrum container for atmospheric transmission as a function of wavelength.

    Parameters
    ----------
    data_file : Name to file containing atmospheric transmission data. Data is assumed to have
        two columns, wavelength and transmission (unscaled dimensionless). If
        this isn't provided, a model is built from a pre-calculated table of values
        from 0.9 to 5.6 microns. The values were generated by the ATRAN model,
        https://ntrs.nasa.gov/citations/19930010877 (Lord, S. D., 1992, NASA
        Technical Memorandum 103957). The extinction is given as a linear transmission
        fraction at an airmass of 1 and 1 mm of precipitable water.

    wave_unit : Units for spectral axis.
    """
    def __init__(
        self,
        data_file: str | Path | None = None,
        wave_unit: u.Unit = u.um,
        **kwargs: str
    ) -> None:
        if data_file is None:
            data_file = download_file(f"{SPECREDUCE_DATA_URL}/extinction/atm_trans_am1.0.dat")
        t = Table.read(data_file, format="ascii", names=['wavelength', 'extinction'])

        # spectral axis is given in microns
        spectral_axis = t['wavelength'].data * wave_unit

        # extinction is given in a dimensionless transmission fraction
        extinction = t['extinction'].data * u.dimensionless_unscaled

        super(AtmosphericTransmission, self).__init__(
            extinction=extinction,
            spectral_axis=spectral_axis,
            **kwargs
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from dataclasses import dataclass, field

import numpy as np
from astropy import units as u
from astropy.modeling import Model, models, fitting
from astropy.nddata import NDData, VarianceUncertainty
from numpy import ndarray
from scipy.integrate import trapezoid
from scipy.interpolate import RectBivariateSpline

from specreduce.compat import Spectrum
from specreduce.core import SpecreduceOperation, ImageLike, MaskingOption
from specreduce.tracing import Trace, FlatTrace

__all__ = ["BoxcarExtract", "HorneExtract", "OptimalExtract"]


def _get_boxcar_weights(center, hwidth, npix):
    """
    Compute weights given an aperture center, half width,
    and number of pixels.

    Based on `get_boxcar_weights()` from a JDAT Notebook by Karl Gordon:
    https://github.com/spacetelescope/jdat_notebooks/blob/main/notebooks/MIRI_LRS_spectral_extraction/miri_lrs_spectral_extraction.ipynb

    Parameters
    ----------
    center : float, required
        The index of the aperture's center pixel on the larger image's
        cross-dispersion axis.

    hwidth : float, required
        Half of the aperture's width in the cross-dispersion direction.

    npix : float, required
        The number of pixels in the larger image's cross-dispersion
        axis.

    Returns
    -------
    weights : `~numpy.ndarray`
        A 2D image with weights assigned to pixels that fall within the
        defined aperture.
    """
    weights = np.zeros((npix))
    if hwidth == 0:
        # the logic below would return all zeros anyways, so might as well save the time
        # (negative widths should be avoided by earlier logic!)
        return weights

    if center - hwidth > npix - 0.5 or center + hwidth < -0.5:
        # entire window is out-of-bounds
        return weights

    lower_edge = max(-0.5, center - hwidth)  # where -0.5 is lower bound of the image
    upper_edge = min(center + hwidth, npix - 0.5)  # where npix-0.5 is upper bound of the image

    # let's avoid recomputing the round repeatedly
    int_round_lower_edge = int(round(lower_edge))
    int_round_upper_edge = int(round(upper_edge))

    # inner pixels that get full weight
    # the round in conjunction with the +1 handles the half-pixel "offset",
    # the upper bound doesn't have the +1 because array slicing is inclusive on the lower index and
    # exclusive on the upper-index
    # NOTE: round(-0.5) == 0, which is helpful here for the case where lower_edge == -0.5
    weights[int_round_lower_edge + 1 : int_round_upper_edge] = 1

    # handle edge pixels (for cases where an edge pixel is fully-weighted, this will set it again,
    # but should still compute a weight of 1.  By using N:N+1, we avoid index errors if the edge
    # is outside the image bounds.  But we do need to avoid negative indices which would count
    # from the end of the array.
    if int_round_lower_edge >= 0:
        weights[int_round_lower_edge : int_round_lower_edge + 1] = (
            round(lower_edge) + 0.5 - lower_edge
        )
    weights[int_round_upper_edge : int_round_upper_edge + 1] = upper_edge - (
        round(upper_edge) - 0.5
    )

    return weights


def _ap_weight_image(trace, width, disp_axis, crossdisp_axis, image_shape):
    """
    Create a weight image that defines the desired extraction aperture.

    Based on `ap_weight_images()` from a JDAT Notebook by Karl Gordon:
    https://github.com/spacetelescope/jdat_notebooks/blob/main/notebooks/MIRI_LRS_spectral_extraction/miri_lrs_spectral_extraction.ipynb

    Parameters
    ----------
    trace : `~specreduce.tracing.Trace`, required
        trace object
    width : float, required
        width of extraction aperture in pixels
    disp_axis : int, required
        dispersion axis
    crossdisp_axis : int, required
        cross-dispersion axis
    image_shape : tuple with 2 elements, required
        size (shape) of image

    Returns
    -------
    wimage : `~numpy.ndarray`
        a 2D weight image defining the aperture
    """
    wimage = np.zeros(image_shape)
    hwidth = 0.5 * width
    image_sizes = image_shape[crossdisp_axis]

    # loop in dispersion direction and compute weights.
    for i in range(image_shape[disp_axis]):
        # TODO trace must handle transposed data (disp_axis == 0)
        # pass trace.trace.data[i] to avoid any mask if part of the regions is out-of-bounds

        # ArrayTrace can have nonfinite or masked data in trace, and this will fail,
        # so figure out how to handle that...

        wimage[:, i] = _get_boxcar_weights(trace.trace.data[i], hwidth, image_sizes)

    return wimage


@dataclass
class BoxcarExtract(SpecreduceOperation):
    """
    Standard boxcar extraction along a trace.

    Example: ::

        trace = FlatTrace(image, trace_pos)
        extract = BoxcarExtract(image, trace)
        spectrum = extract(width=width)


    Parameters
    ----------
    image
        image with 2-D spectral image data
    trace_object
        trace object
    width
        width of extraction aperture in pixels
    disp_axis
        dispersion axis
    crossdisp_axis
        cross-dispersion axis
    mask_treatment
        Specifies how to handle masked or non-finite values in the input image.
        The accepted values are:

        - ``apply``: The image remains unchanged, and any existing mask is combined\
            with a mask derived from non-finite values.
        - ``ignore``: The image remains unchanged, and any existing mask is dropped.
        - ``propagate``: The image remains unchanged, and any masked or non-finite pixel\
            causes the mask to extend across the entire cross-dispersion axis.
        - ``zero_fill``: Pixels that are either masked or non-finite are replaced with 0.0,\
            and the mask is dropped.
        - ``nan_fill``:  Pixels that are either masked or non-finite are replaced with nan,\
            and the mask is dropped.
        - ``apply_mask_only``: The  image and mask are left unmodified.
        - ``apply_nan_only``: The  image is left unmodified, the old mask is dropped, and a\
            new mask is created based on non-finite values.

    Returns
    -------
    spec : `~specutils.Spectrum1D`
        The extracted 1d spectrum expressed in DN and pixel units
    """

    image: ImageLike
    trace_object: Trace
    width: float = 5
    disp_axis: int = 1
    crossdisp_axis: int = 0
    # TODO: should disp_axis and crossdisp_axis be defined in the Trace object?
    mask_treatment: MaskingOption = "apply"
    _valid_mask_treatment_methods = (
        "apply",
        "ignore",
        "propagate",
        "zero_fill",
        "nan_fill",
        "apply_mask_only",
        "apply_nan_only",
    )

    @property
    def spectrum(self):
        return self.__call__()

    def __call__(
        self,
        image: ImageLike | None = None,
        trace: Trace | None = None,
        width: float | None = None,
        disp_axis: int | None = None,
        crossdisp_axis: int | None = None,
    ) -> Spectrum:
        """
        Extract the 1D spectrum using the boxcar method.

        Parameters
        ----------
        image
            The image with 2-D spectral image data
        trace
            The trace object
        width
            The width of extraction aperture in pixels
        disp_axis
            The dispersion axis
        crossdisp_axis
            The cross-dispersion axis

        Returns
        -------
        spec
            The extracted 1d spectrum with flux expressed in the same
            units as the input image, or u.DN, and pixel units
        """
        image = image if image is not None else self.image
        trace = trace or self.trace_object
        width = width or self.width
        disp_axis = disp_axis or self.disp_axis
        cdisp_axis = crossdisp_axis or self.crossdisp_axis

        if width <= 0:
            raise ValueError("The window width must be positive")

        self.image = self._parse_image(
            image, disp_axis=disp_axis, mask_treatment=self.mask_treatment
        )

        # Spectrum extraction
        # ===================
        # Assign no weight to non-finite pixels outside the window. Non-finite pixels inside
        # the window will be propagated to the sum if mask treatment is either ``ignore`` or
        # ``propagate`` or excluded if the chosen mask treatment option is ``apply``. In the
        # latter case, the flux is calculated as the average of the non-masked pixels inside
        # the window multiplied by the window width.
        window_weights = _ap_weight_image(trace, width, disp_axis, cdisp_axis, self.image.shape)

        if self.mask_treatment == "apply":
            image_cleaned = np.where(~self.image.mask, self.image.data * window_weights, 0.0)
            weights = np.where(~self.image.mask, window_weights, 0.0)
            spectrum = (
                image_cleaned.sum(axis=cdisp_axis)
                / weights.sum(axis=cdisp_axis)
                * window_weights.sum(axis=cdisp_axis)
            )
        else:
            image_windowed = np.where(window_weights, self.image.data * window_weights, 0.0)
            spectrum = np.sum(image_windowed, axis=cdisp_axis)

        return Spectrum(spectrum * self.image.unit, spectral_axis=self.image.spectral_axis)


@dataclass
class HorneExtract(SpecreduceOperation):
    """
    Perform a Horne (a.k.a. optimal) extraction on a two-dimensional
    spectrum.

    There are two options for fitting the spatial profile used for
    extraction - by default, a 1D gaussian is fit and as a uniform profile
    across the spectrum. Alternativley, the ``self profile`` option may be
    chosen - when this option is chosen, the spatial profile will be sampled
    (using a default of 10 sample bins, but can be modified with
    ``spatial_profile``) and interpolated between to produce a smoothly varying
    spatial profile across the spectrum.

    If using the Gaussian option for the spatial profile, a background profile
    may be fit (but not subtracted) simultaneously to the data. By default,
    this is done with a 2nd degree polynomial. If using the
    ``interpolated_profile`` option, the background model must be set to None.


    Parameters
    ----------

    image : `~astropy.nddata.NDData`-like or array-like, required
        The input 2D spectrum from which to extract a source. An
        NDData object must specify uncertainty and a mask. An array
        requires use of the ``variance``, ``mask``, & ``unit`` arguments.

    trace_object : `~specreduce.tracing.Trace`, required
        The associated 1D trace object created for the 2D image.

    disp_axis : int, optional
        The index of the image's dispersion axis. [default: 1]

    crossdisp_axis : int, optional
        The index of the image's cross-dispersion axis. [default: 0]

    bkgrd_prof : `~astropy.modeling.Model` or None, optional
        A model for the image's background flux when using the ``gaussian``
        spatial profile. If ``spatial_profile`` is set to ``gaussian``, it defaults
        to ``models.Polynomial1D(2)``. Note that the ``interpolated_profile`` option
        does not support a background model, so ``bkgrd_prof`` must be left as ``None``.

    spatial_profile : str or dict, optional
        The shape of the object profile. The first option is 'gaussian' to fit
        a uniform 1D gaussian to the average of pixels in the cross-dispersion
        direction. The other option is 'interpolated_profile'  - when this
        option is used, the profile is sampled in bins and these samples are
        interpolated between to construct a continuously varying, empirical
        spatial profile for extraction. For this option, if passed in as a
        string (i.e spatial_profile='interpolated_profile') the default values
        for the number of bins used (10) and degree of interpolation
        (linear in x and y, by default) will be used. To set these parameters,
        pass in a dictionary with the keys 'n_bins_interpolated_profile' (which
        accepts an integer number of bins) and 'interp_degree' (which accepts an
        int, or tuple of ints for x and y degree, respectively).
        [default: gaussian]

    variance : `~numpy.ndarray`, optional
        (Only used if ``image`` is not an NDData object.)
        The associated variances for each pixel in the image. Must
        have the same dimensions as ``image``. If all zeros, the variance
        will be ignored and treated as all ones.  If any zeros, those
        elements will be excluded via masking.  If any negative values,
        an error will be raised. [default: None]

    mask : `~numpy.ndarray`, optional
        (Only used if ``image`` is not an NDData object.)
        Whether to mask each pixel in the image. Must have the same
        dimensions as ``image``. If blank, all non-NaN pixels are
        unmasked. [default: None]

    unit : `~astropy.units.Unit` or str, optional
        (Only used if ``image`` is not an NDData object.)
        The associated unit for the data in ``image``. If blank,
        fluxes are interpreted in DN. [default: None]

    """

    image: NDData
    trace_object: Trace
    bkgrd_prof: None | Model = None
    spatial_profile: str | dict = "gaussian"
    variance: np.ndarray = field(default=None)
    mask: np.ndarray = field(default=None)
    unit: np.ndarray = field(default=None)
    disp_axis: int = 1
    crossdisp_axis: int = 0
    # TODO: should disp_axis and crossdisp_axis be defined in the Trace object?

    @property
    def spectrum(self):
        return self.__call__()

    def _parse_image(self, image, variance=None, mask=None, unit=None, disp_axis=1):
        """
        Convert all accepted image types to a consistently formatted
        Spectrum object.

        HorneExtract needs its own version of this method because it is
        more stringent in its requirements for input images. The extra
        arguments are needed to handle cases where these parameters were
        specified as arguments and those where they came as attributes
        of the image object.

        Parameters
        ----------
        image : `~astropy.nddata.NDData`-like or array-like, required
            The image to be parsed. If None, defaults to class' own
            image attribute.
        variance : `~numpy.ndarray`, optional
            (Only used if ``image`` is not an NDData object.)
            The associated variances for each pixel in the image. Must
            have the same dimensions as ``image``. If all zeros, the variance
            will be ignored and treated as all ones.  If any zeros, those
            elements will be excluded via masking.  If any negative values,
            an error will be raised.
        mask : `~numpy.ndarray`, optional
            (Only used if ``image`` is not an NDData object.)
            Whether to mask each pixel in the image. Must have the same
            dimensions as ``image``. If blank, all non-NaN pixels are
            unmasked.
        unit : `~astropy.units.Unit` or str, optional
            (Only used if ``image`` is not an NDData object.)
            The associated unit for the data in ``image``. If blank,
            fluxes are interpreted in DN.
        disp_axis : int, optional
            The index of the image's dispersion axis. Should not be
            changed until operations can handle variable image
            orientations. [default: 1]
        """

        if isinstance(image, np.ndarray):
            img = image
        elif isinstance(image, u.quantity.Quantity):
            img = image.value
        else:  # NDData, including CCDData and Spectrum
            img = image.data

        # mask is set as None when not specified upon creating a Spectrum
        # object, so we must check whether it is absent *and* whether it's
        # present but set as None
        if getattr(image, "mask", None) is not None:
            mask = image.mask
        elif mask is not None:
            pass
        else:
            # if user provides no mask at all, don't mask anywhere
            mask = np.zeros_like(img)

        if img.shape != mask.shape:
            raise ValueError("image and mask shapes must match.")

        # Process uncertainties, converting to variances when able and throwing
        # an error when uncertainties are missing or less easily converted
        if hasattr(image, "uncertainty") and image.uncertainty is not None:
            if image.uncertainty.uncertainty_type == "var":
                variance = image.uncertainty.array
            elif image.uncertainty.uncertainty_type == "std":
                warnings.warn(
                    "image NDData object's uncertainty "
                    "interpreted as standard deviation. if "
                    "incorrect, use VarianceUncertainty when "
                    "assigning image object's uncertainty."
                )
                variance = image.uncertainty.array**2
            elif image.uncertainty.uncertainty_type == "ivar":
                variance = 1 / image.uncertainty.array
            else:
                # other options are InverseUncertainty and UnknownUncertainty
                raise ValueError(
                    "image NDData object has unexpected "
                    "uncertainty type. instead, try "
                    "VarianceUncertainty or StdDevUncertainty."
                )
        elif hasattr(image, "uncertainty") and image.uncertainty is None:
            # ignore variance arg to focus on updating NDData object
            raise ValueError("image NDData object lacks uncertainty")
        else:
            if variance is None:
                raise ValueError(
                    "if image is a numpy or Quantity array, a "
                    "variance must be specified. consider "
                    "wrapping it into one object by instead "
                    "passing an NDData image."
                )
            elif image.shape != variance.shape:
                raise ValueError("image and variance shapes must match")

        if np.any(variance < 0):
            raise ValueError("variance must be fully positive")
        if np.all(variance == 0):
            # technically would result in infinities, but since they're all
            # zeros, we can override ones to simulate an unweighted case
            variance = np.ones_like(variance)
        if np.any(variance == 0):
            # exclude such elements by editing the input mask
            mask[variance == 0] = True
            # replace the variances to avoid a divide by zero warning
            variance[variance == 0] = np.nan

        variance = VarianceUncertainty(variance)

        unit = getattr(image, "unit", u.Unit(unit) if unit is not None else u.Unit("DN"))

        spectral_axis = getattr(image, "spectral_axis", np.arange(img.shape[disp_axis]) * u.pix)

        return Spectrum(img * unit, spectral_axis=spectral_axis, uncertainty=variance, mask=mask)

    def _fit_gaussian_spatial_profile(
        self, img: ndarray, disp_axis: int, crossdisp_axis: int, or_mask: ndarray, bkgrd_prof: Model
    ):
        """Fit a 1D Gaussian spatial profile to spectrum in `img`.

        Fits an 1D Gaussian profile to spectrum in `img`. Takes the weighted mean
        of  ``img`` along the cross-dispersion axis all ignoring masked pixels
        (i.e, takes the mean of each row for a horizontal trace). A Background model
        (optional) is fit simultaneously. Returns an `astropy.model.Gaussian1D` (or
        compound model, if `bkgrd_prof` is supplied) fit to data.
        """
        nrows = img.shape[crossdisp_axis]
        xd_pixels = np.arange(nrows)

        # co-add signal in each image row, ignore masked pixels
        coadd = np.ma.masked_array(img, mask=or_mask).mean(disp_axis)

        # use the sum of brightest row as an inital guess for Gaussian amplitude,
        # the the location of the brightest row as an initial guess for the mean
        gauss_prof = models.Gaussian1D(amplitude=coadd.max(), mean=coadd.argmax(), stddev=2)

        # Fit extraction kernel (Gaussian + background model) to coadded rows
        # with combined model (must exclude masked indices manually;
        # LevMarLSQFitter does not)
        if bkgrd_prof is not None:
            ext_prof = gauss_prof + bkgrd_prof
        else:
            # add a trivial constant model so attribute names are the same
            ext_prof = gauss_prof + models.Const1D(0, fixed={"amplitude": True})

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fitter = fitting.LMLSQFitter()
            fit_ext_kernel = fitter(ext_prof, xd_pixels[~coadd.mask], coadd.compressed())
        return fit_ext_kernel

    def _fit_spatial_profile(
        self,
        img: ndarray,
        disp_axis: int,
        crossdisp_axis: int,
        mask: ndarray,
        n_bins: int,
        kx: int,
        ky: int,
    ) -> RectBivariateSpline:
        """
        Fit a spatial profile by sampling the median profile along the dispersion direction.

        This method extracts the spatial profile from an input spectrum by binning
        the data along the dispersion axis. It calculates the median profile for each bin,
        normalizes it, and then interpolates between these profiles to create a smooth
        2D representation of the spatial profile. The resulting interpolator object can be
        used to evaluate the spatial profile at any coordinate within the bounds of the data.

        Parameters
        ----------
        img
            The 2D array of spectral data to process.
        disp_axis
            The image axis corresponding to the dispersion direction.
        crossdisp_axis
            The image axis corresponding to the cross-dispersion direction.
        mask
            A boolean mask array with the same shape as the image. Values of ``True``
            in the mask indicate invalid data points to be ignored during computation.
        n_bins
            The number of bins to use along the dispersion axis for sampling
            the median spatial profile.
        kx
            The degree of the spline along the dispersion axis.
        ky
            The degree of the spline along the cross-dispersion axis.

        Returns
        -------
        RectBivariateSpline
            Interpolator object that provides a smoothed 2D spatial profile.
        """
        img = np.where(~mask, img, np.nan)
        nrows = img.shape[crossdisp_axis]
        ncols = img.shape[disp_axis]
        samples = np.zeros((n_bins, nrows))

        sample_locs = np.linspace(0, ncols - 1, n_bins + 1, dtype=int)
        bin_centers = [
            (sample_locs[i] + sample_locs[i + 1]) // 2 for i in range(len(sample_locs) - 1)
        ]

        for i in range(n_bins):
            bin_median = np.nanmedian(img[:, sample_locs[i] : sample_locs[i + 1]], axis=disp_axis)
            samples[i, :] = bin_median / bin_median.sum()

        return RectBivariateSpline(x=bin_centers, y=np.arange(nrows), z=samples, kx=kx, ky=ky)

    def __call__(
        self,
        image=None,
        trace_object=None,
        disp_axis=None,
        crossdisp_axis=None,
        bkgrd_prof=None,
        spatial_profile=None,
        n_bins_interpolated_profile=None,
        interp_degree_interpolated_profile=None,
        variance=None,
        mask=None,
        unit=None,
    ):
        """
        Run the Horne calculation on a region of an image and extract a 1D spectrum.

        Parameters
        ----------

        image : `~astropy.nddata.NDData`-like or array-like, required
            The input 2D spectrum from which to extract a source. An
            NDData object must specify uncertainty and a mask. An array
            requires use of the ``variance``, ``mask``, & ``unit`` arguments.

        trace_object : `~specreduce.tracing.Trace`, required
            The associated 1D trace object created for the 2D image.

        disp_axis : int, optional
            The index of the image's dispersion axis.

        crossdisp_axis : int, optional
            The index of the image's cross-dispersion axis.

        bkgrd_prof : `~astropy.modeling.Model`, optional
            A model for the image's background flux when using the ``gaussian``
            spatial profile. If ``spatial_profile`` is set to ``gaussian``, it defaults
            to ``models.Polynomial1D(2)``. Note that the ``interpolated_profile`` option
            does not support a background model, so ``bkgrd_prof`` must be left as ``None``.

        spatial_profile : str or dict, optional
            The shape of the object profile. The first option is 'gaussian' to fit
            a uniform 1D gaussian to the average of pixels in the cross-dispersion
            direction. The other option is 'interpolated_profile'  - when this
            option is used, the profile is sampled in bins and these samples are
            interpolated between to construct a continuously varying, empirical
            spatial profile for extraction. For this option, if passed in as a
            string (i.e spatial_profile='interpolated_profile') the default values
            for the number of bins used (10) and degree of interpolation
            (linear in x and y, by default) will be used. To set these parameters,
            pass in a dictionary with the keys 'n_bins_interpolated_profile' (which
            accepts an integer number of bins) and 'interp_degree' (which accepts an
            int, or tuple of ints for x and y degree, respectively).
            [default: gaussian]

        variance : `~numpy.ndarray`, optional
            (Only used if ``image`` is not an NDData object.)
            The associated variances for each pixel in the image. Must
            have the same dimensions as ``image``. If all zeros, the variance
            will be ignored and treated as all ones.  If any zeros, those
            elements will be excluded via masking.  If any negative values,
            an error will be raised.

        mask : `~numpy.ndarray`, optional
            (Only used if ``image`` is not an NDData object.)
            Whether to mask each pixel in the image. Must have the same
            dimensions as ``image``. If blank, all non-NaN pixels are
            unmasked.

        unit : `~astropy.units.Unit` or str, optional
            (Only used if ``image`` is not an NDData object.)
            The associated unit for the data in ``image``. If blank,
            fluxes are interpreted in DN.


        Returns
        -------
        spec_1d : `~specutils.Spectrum1D`
            The final, Horne extracted 1D spectrum.
        """
        image = image if image is not None else self.image
        trace_object = trace_object if trace_object is not None else self.trace_object
        disp_axis = disp_axis if disp_axis is not None else self.disp_axis
        crossdisp_axis = crossdisp_axis if crossdisp_axis is not None else self.crossdisp_axis
        bkgrd_prof = bkgrd_prof if bkgrd_prof is not None else self.bkgrd_prof
        profile = spatial_profile if spatial_profile is not None else self.spatial_profile
        variance = variance if variance is not None else self.variance
        mask = mask if mask is not None else self.mask
        unit = unit if unit is not None else self.unit

        profile_choices = ("gaussian", "interpolated_profile")

        if not isinstance(profile, (str, dict)):
            raise ValueError("spatial_profile must be a string or dictionary.")
        if isinstance(profile, str):
            profile = dict(name=profile)

        profile_type = profile["name"].lower()
        if profile_type not in profile_choices:
            raise ValueError("spatial_profile must be one of" f"{', '.join(profile_choices)}")

        n_bins_interpolated_profile = profile.get("n_bins_interpolated_profile", 10)
        interp_degree_interpolated_profile = profile.get("interp_degree_interpolated_profile", 1)

        if bkgrd_prof is None and profile_type == 'gaussian':
            bkgrd_prof = models.Polynomial1D(2)

        self.image = self._parse_image(image, variance, mask, unit, disp_axis)
        variance = self.image.uncertainty.represent_as(VarianceUncertainty).array
        mask = self.image.mask.astype(bool) | (~np.isfinite(self.image.data))
        unit = self.image.unit
        img = self.image.data

        ncross = img.shape[crossdisp_axis]
        ndisp = img.shape[disp_axis]

        # If the trace is not flat, shift the rows in each column
        # so the image is aligned along the trace:
        if not isinstance(trace_object, FlatTrace):
            img = _align_along_trace(
                img, trace_object.trace, disp_axis=disp_axis, crossdisp_axis=crossdisp_axis
            )

        if profile_type == "gaussian":
            fit_ext_kernel = self._fit_gaussian_spatial_profile(
                img, disp_axis, crossdisp_axis, mask, bkgrd_prof
            )
            if isinstance(trace_object, FlatTrace):
                mean_cross_pix = trace_object.trace
            else:
                mean_cross_pix = np.broadcast_to(ncross // 2, ndisp)
        else:  # interpolated_profile
            # determine interpolation degree from input and make tuple if int
            # this can also be moved to another method to parse the input
            # 'spatial_profile' arg, eventually
            if isinstance(interp_degree_interpolated_profile, int):
                kx = ky = interp_degree_interpolated_profile
            else:  # if input is tuple of ints
                if not isinstance(interp_degree_interpolated_profile, tuple):
                    raise ValueError(
                        "``interp_degree_interpolated_profile`` must be ",
                        "an integer or tuple of integers.",
                    )
                if not all(isinstance(x, int) for x in interp_degree_interpolated_profile):
                    raise ValueError(
                        "``interp_degree_interpolated_profile`` must be ",
                        "an integer or tuple of integers.",
                    )
                kx, ky = interp_degree_interpolated_profile

            interp_spatial_prof = self._fit_spatial_profile(
                img, disp_axis, crossdisp_axis, mask, n_bins_interpolated_profile, kx, ky
            )

            # add private attribute to save fit profile. should this be public?
            self._interp_spatial_prof = interp_spatial_prof

        xd_pixels = np.arange(ncross)
        kernel_vals = np.zeros(img.shape)
        norms = np.full(ndisp, np.nan)
        valid = ~mask

        if profile_type == "gaussian":
            norms[:] = fit_ext_kernel.amplitude_0 * fit_ext_kernel.stddev_0 * np.sqrt(2 * np.pi)

        for idisp in range(ndisp):
            if not np.any(valid[:, idisp]):
                continue
            if profile_type == "gaussian":
                fit_ext_kernel.mean_0 = mean_cross_pix[idisp]
                fitted_col = fit_ext_kernel(xd_pixels)
                kernel_vals[:, idisp] = fitted_col
            else:
                fitted_col = interp_spatial_prof(idisp, xd_pixels)
                kernel_vals[:, idisp] = fitted_col
                norms[idisp] = trapezoid(fitted_col, dx=1)[0]

        with np.errstate(divide="ignore", invalid="ignore"):
            num = np.sum(np.where(valid, img * kernel_vals / variance, 0.0), axis=crossdisp_axis)
            den = np.sum(np.where(valid, kernel_vals**2 / variance, 0.0), axis=crossdisp_axis)
            extraction = (num / den) * norms

        return Spectrum(extraction * unit, spectral_axis=self.image.spectral_axis)


def _align_along_trace(img, trace_array, disp_axis=1, crossdisp_axis=0):
    """
    Given an arbitrary trace ``trace_array`` (an np.ndarray), roll
    all columns of ``nddata`` to shift the NDData's pixels nearest
    to the trace to the center of the spatial dimension of the
    NDData.
    """
    # TODO: this workflow does not support extraction for >2D spectra
    if not (disp_axis == 1 and crossdisp_axis == 0):
        # take the transpose to ensure the rows are the cross-disp axis:
        img = img.T

    n_rows, n_cols = img.shape

    # indices of all columns, in their original order
    rows = np.broadcast_to(np.arange(n_rows)[:, None], img.shape)
    cols = np.broadcast_to(np.arange(n_cols), img.shape)

    # we want to "roll" each column so that the trace sits in
    # the central row of the final image
    shifts = trace_array.astype(int) - n_rows // 2

    # we wrap the indices so we don't index out of bounds
    shifted_rows = np.mod(rows + shifts[None, :], n_rows)

    return img[shifted_rows, cols]


@dataclass
class OptimalExtract(HorneExtract):
    """
    An alias for `HorneExtract`.
    """

    __doc__ += HorneExtract.__doc__
    pass



from functools import cached_property
from typing import Callable

import astropy.units as u
import gwcs
import numpy as np
from astropy.modeling import models, CompoundModel
from astropy.nddata import VarianceUncertainty
from gwcs import coordinate_frames
from numpy.ma import MaskedArray
from numpy.typing import ArrayLike, NDArray
from scipy.interpolate import interp1d

from specreduce.compat import Spectrum


__all__ = ["WavelengthSolution1D"]


def _diff_poly1d(m: models.Polynomial1D) -> models.Polynomial1D:
    """Compute the derivative of a Polynomial1D model.

    Computes the derivative of a Polynomial1D model and returns a new Polynomial1D
    model representing the derivative. The coefficients of the input model are
    used to calculate the coefficients of the derivative model. For a Polynomial1D
    of degree n, the derivative is a Polynomial1D of degree n-1.

    Parameters
    ----------
    m
        A Polynomial1D model for which the derivative is to be computed.

    Returns
    -------
    A new Polynomial1D model representing the derivative of the input Polynomial1D model.
    """
    coeffs = {f"c{i-1}": i * getattr(m, f"c{i}").value for i in range(1, m.degree + 1)}
    return models.Polynomial1D(m.degree - 1, **coeffs)


class WavelengthSolution1D:
    def __init__(
        self,
        p2w: None | CompoundModel,
        bounds_pix: tuple[int, int],
        unit: u.Unit,
    ) -> None:
        """Class defining a one-dimensional wavelength solution.

        This class manages the mapping between pixel positions and wavelength values in a 1D
        spectrum, supporting both forward and reverse transformations. It provides methods for
        resampling spectra in the pixel-to-wavelength space while conserving flux, and integrates
        with GWCS for coordinate transformations.

        Initializes an object with pixel-to-wavelength transformation, pixel bounds, and
        measurement unit. Also, converts the unit to its LaTeX string representation.

        Parameters
        ----------
        p2w
            The pixel-to-wavelength transformation model. If None, no transformation
            will be set.
        bounds_pix
            The lower and upper pixel bounds defining the range of the spectrum.
        unit
            The wavelength unit.
        """
        self.unit = unit
        self._unit_str = unit.to_string("latex")
        self.bounds_pix: tuple[int, int] = bounds_pix
        self.bounds_wav: tuple[float, float] | None = None
        self._p2w: None | CompoundModel = None
        self.p2w = p2w

    @property
    def p2w(self) -> None | CompoundModel:
        """Pixel-to-wavelength transformation."""
        return self._p2w

    @p2w.setter
    def p2w(self, m: CompoundModel) -> None:
        self._p2w = m
        self.ref_pixel = m[0].offset.value if m is not None else None

        if "p2w_dldx" in self.__dict__:
            del self.p2w_dldx
        if "w2p" in self.__dict__:
            del self.w2p
        if "gwcs" in self.__dict__:
            del self.gwcs

    @cached_property
    def p2w_dldx(self) -> CompoundModel:
        """Partial derivative of the pixel-to-wavelength transformation, (d lambda) / (d pix)."""
        return models.Shift(self._p2w.offset_0) | _diff_poly1d(self._p2w[1])

    @cached_property
    def w2p(self) -> Callable:
        """Wavelength-to-pixel transformation."""
        p = np.arange(self.bounds_pix[0] - 2, self.bounds_pix[1] + 2)
        self.bounds_wav = self.p2w(self.bounds_pix)
        return interp1d(self.p2w(p), p, bounds_error=False, fill_value=np.nan)

    def pix_to_wav(self, pix: float | ArrayLike) -> float | NDArray | MaskedArray:
        """Map pixel values into wavelength values.

        Parameters
        ----------
        pix
            The pixel value(s) to be transformed into wavelength value(s).

        Returns
        -------
        Transformed wavelength value(s) corresponding to the input pixel value(s).
        """
        if isinstance(pix, MaskedArray):
            wav = self.p2w(pix.data)
            return np.ma.masked_array(wav, mask=pix.mask)
        else:
            return self.p2w(pix)

    def wav_to_pix(self, wav: float | ArrayLike) -> float | NDArray | MaskedArray:
        """Map wavelength values into pixel values.

        Parameters
        ----------
        wav
            The wavelength value(s) to be converted into pixel value(s).

        Returns
        -------
        The corresponding pixel value(s) for the input wavelength(s).
        """
        if isinstance(wav, MaskedArray):
            pix = self.w2p(wav.data)
            return np.ma.masked_array(pix, mask=wav.mask)
        else:
            return self.w2p(wav)

    @cached_property
    def gwcs(self) -> gwcs.wcs.WCS:
        """GWCS object defining the mapping between pixel and spectral coordinate frames."""
        pixel_frame = coordinate_frames.CoordinateFrame(
            1, "SPECTRAL", (0,), axes_names=["x"], unit=[u.pix]
        )
        spectral_frame = coordinate_frames.SpectralFrame(
            axes_names=("wavelength",), unit=[self.unit]
        )
        pipeline = [(pixel_frame, self._p2w), (spectral_frame, None)]
        return gwcs.wcs.WCS(pipeline)

    def resample(
        self,
        spectrum: "Spectrum",
        nbins: int | None = None,
        wlbounds: tuple[float, float] | None = None,
        bin_edges: ArrayLike | None = None,
    ) -> Spectrum:
        """Bin the given pixel-space 1D spectrum to a wavelength space conserving the flux.

        This method bins a pixel-space spectrum to a wavelength space using the computed
        pixel-to-wavelength and wavelength-to-pixel transformations and their derivatives with
        respect to the spectral axis. The binning is exact and conserves the integrated flux.

        Parameters
        ----------
        spectrum
            A Spectrum instance containing the flux to be resampled over the wavelength
            space.

        nbins
            The number of bins for resampling. If not provided, it defaults to the size of the
            input spectrum.

        wlbounds
            A tuple specifying the starting and ending wavelengths for resampling. If not
            provided, the wavelength bounds are inferred from the object's methods and the
            entire flux array is used.

        bin_edges
            Explicit bin edges in the wavelength space. Should be an 1D array-like [e_0, e_1,
            ..., e_n] with n = nbins + 1. The bins are created as [[e_0, e_1], [e_1, e_2], ...,
            [e_n-1, n]]. If provided, ``nbins`` and ``wlbounds`` are ignored.

        Returns
        -------
            1D spectrum binned to the specified wavelength bins.
        """
        if nbins is not None and nbins < 0:
            raise ValueError("Number of bins must be non-zero and positive.")

        if self._p2w is None:
            raise ValueError("Wavelength solution not set.")

        flux = spectrum.flux.value
        pixels = spectrum.spectral_axis.value

        if spectrum.uncertainty is not None:
            ucty = spectrum.uncertainty.represent_as(VarianceUncertainty).array
            ucty_type = type(spectrum.uncertainty)
        else:
            ucty = np.zeros_like(flux)
            ucty_type = VarianceUncertainty
        npix = flux.size
        nbins = npix if nbins is None else nbins
        if wlbounds is None:
            l1, l2 = self.p2w(pixels[[0, -1]] + np.array([-0.5, 0.5]))
        else:
            l1, l2 = wlbounds

        if bin_edges is not None:
            bin_edges_wav = np.asarray(bin_edges)
            nbins = bin_edges_wav.size - 1
        else:
            bin_edges_wav = np.linspace(l1, l2, num=nbins + 1)

        bin_edges_pix = np.clip(self.w2p(bin_edges_wav) + 0.5, 0, npix - 1e-12)
        bin_edge_ix = np.floor(bin_edges_pix).astype(int)
        bin_edge_w = bin_edges_pix - bin_edge_ix
        bin_centers_wav = 0.5 * (bin_edges_wav[:-1] + bin_edges_wav[1:])
        flux_wl = np.zeros(nbins)
        ucty_wl = np.zeros(nbins)
        weights = np.zeros(npix)

        dldx = np.diff(self.p2w(np.arange(pixels[0], pixels[-1] + 2) - 0.5))

        for i in range(nbins):
            i1, i2 = bin_edge_ix[i : i + 2]
            weights[:] = 0.0
            if i1 != i2:
                weights[i1 + 1 : i2] = 1.0
                weights[i1] = 1 - bin_edge_w[i]
                weights[i2] = bin_edge_w[i + 1]
                sl = slice(i1, i2 + 1)
                w = weights[sl]
                flux_wl[i] = (w * flux[sl] * dldx[sl]).sum()
                ucty_wl[i] = (w**2 * ucty[sl] * dldx[sl]).sum()
            else:
                fracw = bin_edges_pix[i + 1] - bin_edges_pix[i]
                flux_wl[i] = fracw * flux[i1] * dldx[i1]
                ucty_wl[i] = fracw**2 * ucty[i1] * dldx[i1]

        bin_widths_wav = np.diff(bin_edges_wav)
        flux_wl = flux_wl / bin_widths_wav * spectrum.flux.unit / self.unit
        ucty_wl = VarianceUncertainty(ucty_wl / bin_widths_wav**2).represent_as(ucty_type)
        return Spectrum(flux_wl, bin_centers_wav * self.unit, uncertainty=ucty_wl)
